from __future__ import annotations

import typing

from ctc import spec
from .. import abi_utils
from .. import binary_utils

if typing.TYPE_CHECKING:
    from typing_extensions import Literal
    import polars as pl


def get_event_df_columns(
    binary_format: Literal['binary', 'prefix_hex']
) -> typing.Sequence[tuple[str, pl.datatypes.DataTypeClass]]:
    import polars as pl

    if binary_format == 'binary':
        binary: pl.datatypes.DataTypeClass = pl.datatypes.Binary
    elif binary_format in ['prefix_hex', 'raw_hex']:
        binary = pl.datatypes.Utf8
    else:
        raise Exception('unknown binary format: ' + str(binary_format))

    return [
        ('block_number', pl.datatypes.Int64),
        ('transaction_index', pl.datatypes.Int64),
        ('log_index', pl.datatypes.Int64),
        ('transaction_hash', binary),
        ('contract_address', binary),
        ('event_hash', binary),
        ('topic1', binary),
        ('topic2', binary),
        ('topic3', binary),
        ('unindexed', binary),
    ]


def _parse_event_query_type(
    *,
    contract_address: spec.Address | None = None,
    event_hash: typing.Any | None = None,
    topic1: typing.Any | None = None,
    topic2: typing.Any | None = None,
    topic3: typing.Any | None = None,
) -> int:

    # parse input flags
    has_contract = contract_address is not None
    has_event_type = event_hash is not None
    has_other_topics = (
        topic1 is not None or topic2 is not None or topic3 is not None
    )

    # return type based on input flags
    if has_contract and has_event_type and not has_other_topics:
        return 1
    elif not has_contract and has_event_type and not has_other_topics:
        return 2
    elif has_contract and not has_event_type and not has_other_topics:
        return 3
    elif not has_contract and not has_event_type and not has_other_topics:
        return 4
    elif has_contract and has_event_type and has_other_topics:
        return 5
    elif not has_contract and has_event_type and has_other_topics:
        return 6
    elif has_contract and not has_event_type and has_other_topics:
        return 7
    elif not has_contract and not has_event_type and has_other_topics:
        return 8
    else:
        raise Exception('could not parse query type')


def _is_topic_binary(topic: typing.Any) -> bool:
    return isinstance(topic, bytes) or spec.is_hex_data(topic)


async def _async_parse_event_query_args(
    contract_address: spec.Address | None,
    *,
    event_name: str | None,
    event_abi: spec.EventABI | None,
    event_hash: str | None,
    named_topics: typing.Mapping[str, typing.Any] | None,
    topic1: typing.Any | None = None,
    topic2: typing.Any | None = None,
    topic3: typing.Any | None = None,
    topic1_is_binary: bool | None = None,
    topic2_is_binary: bool | None = None,
    topic3_is_binary: bool | None = None,
    context: spec.Context,
) -> tuple[typing.Sequence[bytes | None], spec.EventABI | None]:
    """compute encoded topics and event abi of query as needed"""

    # obtain event abi if needed
    if event_hash is not None and event_abi is not None:
        if event_hash != abi_utils.get_event_hash(event_abi):
            raise Exception('conflicting event hashes for given inputs')
    if topic1_is_binary is None:
        topic1_is_binary = _is_topic_binary(topic1)
    if topic2_is_binary is None:
        topic2_is_binary = _is_topic_binary(topic2)
    if topic3_is_binary is None:
        topic3_is_binary = _is_topic_binary(topic3)
    topics_decoded = (
        (topic1 is not None and not topic1_is_binary)
        or (topic2 is not None and not topic1_is_binary)
        or (topic3 is not None and not topic1_is_binary)
    )
    using_event_name = event_hash is None and event_name is not None
    if event_abi is None:
        if named_topics is not None or topics_decoded or using_event_name:
            event_abi = await abi_utils.async_get_event_abi(
                contract_address=contract_address,
                event_name=event_name,
                event_hash=event_hash,
                context=context,
            )

    # get encoded event_abi_hash
    if event_abi is not None:
        event_hash = abi_utils.get_event_hash(event_abi)
    if event_hash is not None:
        encoded_event_hash = binary_utils.to_binary(event_hash)
    else:
        encoded_event_hash = None

    # parse named event topics
    if named_topics is not None:
        if event_abi is None:
            raise Exception('could not obtain event_abi')
        names = abi_utils.get_event_indexed_names(event_abi)
        if len(names) >= 1 and names[0] in named_topics:
            topic1 = named_topics[names[0]]
            topic1_is_binary = False
        if len(names) >= 2 and names[1] in named_topics:
            topic2 = named_topics[names[1]]
            topic2_is_binary = False
        if len(names) >= 3 and names[2] in named_topics:
            topic3 = named_topics[names[2]]
            topic3_is_binary = False
        for name in named_topics.keys():
            if name not in names:
                raise Exception('unknown topic name: ' + str(name))

    # encode events
    if event_abi is not None:
        indexed_types = abi_utils.get_event_indexed_types(event_abi)
    if topic1 is not None:
        if not topic1_is_binary:
            topic1 = abi_utils.abi_encode(topic1, indexed_types[0])
        else:
            topic1 = binary_utils.to_binary(topic1, n_bytes=32)
    if topic2 is not None:
        if not topic2_is_binary:
            topic2 = abi_utils.abi_encode(topic2, indexed_types[1])
        else:
            topic2 = binary_utils.to_binary(topic2, n_bytes=32)
    if topic3 is not None:
        if not topic3_is_binary:
            topic3 = abi_utils.abi_encode(topic3, indexed_types[2])
        else:
            topic3 = binary_utils.to_binary(topic3, n_bytes=32)

    # return result
    return (encoded_event_hash, topic1, topic2, topic3), event_abi


async def async_scrub_db_queries(
    *,
    contract_address: spec.Address | None,
    event_hash: str | None = None,
    encoded_topic1: spec.BinaryData | None = None,
    encoded_topic2: spec.BinaryData | None = None,
    encoded_topic3: spec.BinaryData | None = None,
    start_block: spec.BlockNumberReference | None = None,
    end_block: spec.BlockNumberReference | None = None,
    context: spec.Context,
) -> None:
    """join contiguous and overlapping event query records together"""

    import toolsql

    from ctc import config
    from ctc import db
    from ctc.toolbox import range_utils

    # query queries
    queries = await db.async_query_event_queries(
        contract_address=contract_address,
        event_hash=event_hash,
        topic1=encoded_topic1,
        topic2=encoded_topic2,
        topic3=encoded_topic3,
        start_block=start_block,
        end_block=end_block,
        context=context,
    )
    if queries is None:
        queries = []

    # sort queries into groups
    if typing.TYPE_CHECKING:
        QueryTuple = typing.Tuple[
            int,
            typing.Union[spec.Address, None],
            typing.Union[spec.BinaryData, None],
            typing.Union[spec.BinaryData, None],
            typing.Union[spec.BinaryData, None],
            typing.Union[spec.BinaryData, None],
        ]
    query_sets: typing.MutableMapping[
        QueryTuple, typing.MutableSequence[tuple[int, int]]
    ]
    query_sets = {}
    query_ids: typing.MutableMapping[
        QueryTuple, typing.MutableMapping[tuple[int, int], int]
    ]
    query_ids = {}
    for query in queries:
        if query is None:
            continue
        query_set: QueryTuple = (
            query['query_type'],
            query['contract_address'],
            query['event_hash'],
            query['topic1'],
            query['topic2'],
            query['topic3'],
        )
        query_range = (query['start_block'], query['end_block'])
        query_sets.setdefault(query_set, [])
        query_sets[query_set].append(query_range)
        query_ids.setdefault(query_set, {})
        query_ids[query_set][query_range] = query['query_id']

    # join contiguous and overlapping queries
    to_delete: typing.MutableSequence[int] = []
    to_insert: typing.MutableSequence[spec.EventQuery] = []
    for query_set, query_ranges in query_sets.items():
        scrubbed_lists = range_utils.combine_overlapping_ranges(query_ranges)
        scrubbed = [(qr[0], qr[1]) for qr in scrubbed_lists]
        if len(scrubbed) != len(query_ranges):
            for query_range in query_ranges:
                if query_range not in scrubbed:
                    to_delete.append(query_ids[query_set][query_range])
            for query_range in scrubbed:
                if query_range not in query_ranges:
                    new_query: spec.DBEventQuery = {
                        'query_type': query_set[0],
                        'contract_address': query_set[1],
                        'event_hash': query_set[2],
                        'topic1': query_set[3],
                        'topic2': query_set[4],
                        'topic3': query_set[5],
                        'start_block': query_range[0],
                        'end_block': query_range[1],
                    }
                    to_insert.append(new_query)

    # perform db operations
    db_config = config.get_context_db_config(
        schema_name='events',
        context=context,
    )
    async with toolsql.async_connect(db_config) as conn:
        await db.async_delete_event_queries(
            query_ids=to_delete,
            context=context,
            conn=conn,
        )
        await db.async_upsert_event_queries(
            event_queries=to_insert,
            context=context,
            conn=conn,
        )


def print_event_query_summary(
    *,
    contract_address: spec.Address | None,
    event_hash: typing.Any | None,
    topic1: typing.Any | None,
    topic2: typing.Any | None,
    topic3: typing.Any | None,
    start_block: int,
    end_block: int,
) -> None:

    from ctc import cli

    cli.print_bullet(key='contract_address', value=contract_address)
    if event_hash is not None:
        event_hash = binary_utils.to_hex(event_hash)
    cli.print_bullet(key='event_hash', value=event_hash)
    cli.print_bullet(key='topic1', value=topic1)
    cli.print_bullet(key='topic2', value=topic2)
    cli.print_bullet(key='topic3', value=topic3)
    cli.print_bullet(key='start_block', value=start_block)
    cli.print_bullet(key='end_block', value=end_block)




# MOVE create_missing_schemas TO dba_utils

from __future__ import annotations

import os
import typing

import toolsql

from ctc import config
from ctc import spec
from .management import dba_utils
from .management import version_utils


def connect(
    context: spec.Context,
    schema: spec.SchemaName,
    *,
    read_only: bool = False,
    db_config: toolsql.DBConfig | None = None,
) -> toolsql.Connection:
    if db_config is None:
        db_config = config.get_context_db_config(
            context=context, schema_name=schema
        )
    if read_only:
        db_config = db_config.copy()
        db_config['driver'] = 'connectorx'
    return toolsql.connect(db_config)


def async_connect(
    context: spec.Context,
    schema: spec.SchemaName,
    *,
    read_only: bool = False,
    db_config: toolsql.DBConfig | None = None,
) -> toolsql.AsyncConnection:
    if db_config is None:
        db_config = config.get_context_db_config(
            context=context, schema_name=schema
        )
    if read_only:
        db_config = db_config.copy()
        db_config['driver'] = 'connectorx'
    return toolsql.async_connect(db_config)


def create_missing_schemas(
    schema_name: spec.SchemaName | None = None,
    *,
    schema_names: typing.Sequence[spec.SchemaName] | None = None,
    context: spec.Context = None,
) -> None:

    if schema_name is not None and schema_names is not None:
        raise Exception('specify only one of schema_name or schema_names')
    if schema_name is None:
        if schema_names is None or len(schema_names) == 0:
            raise Exception('must specify schema_name or schema_names')
        else:
            schema_name = schema_names[0]
    if schema_names is None:
        if schema_name is None:
            raise Exception('must specify schema_name or schema_names')
        else:
            schema_names = [schema_name]

    # get db config
    db_config = config.get_context_db_config(
        context=context, schema_name=schema_name
    )

    # create directory if need be
    if db_config['dbms'] == 'sqlite':
        path = db_config['path']
        if path is None:
            raise Exception('sqlite path not specified')
        pathdir = os.path.dirname(os.path.abspath(path))
        if pathdir is None:
            raise Exception('invalid path dir')
        os.makedirs(pathdir, exist_ok=True)

    # create missing tables
    network = config.get_context_chain_id(context)
    with toolsql.connect(db_config) as conn:

        # check that schema versions being tracked
        if not version_utils.is_schema_versions_initialized(conn=conn):
            dba_utils.initialize_schema_versions(conn=conn)

        for schema_name in schema_names:

            # check if schema in database
            schema_version = version_utils.get_schema_version(
                schema_name=schema_name,
                context=dict(network=network),
                conn=conn,
            )

            # create schema if missing
            if schema_version is None:
                dba_utils.initialize_schema(
                    schema_name=schema_name,
                    context=context,
                    conn=conn,
                )




from __future__ import annotations

from ctc import spec

from ... import contract_utils
from . import contract_abi_modification


async def async_get_contract_abi(
    contract_address: spec.Address,
    *,
    block: spec.BlockNumberReference | None = None,
    proxy_implementation: spec.Address | None = None,
    verbose: bool = True,
    context: spec.Context = None,
) -> spec.ContractABI:
    """retrieve abi of contract either from local database or block explorer

    for addresses that change ABI's over time, use db_query=False to skip cache
    """
    from ctc import config

    # load from db
    read_cache, write_cache = config.get_context_cache_read_write(
        context=context, schema_name='contract_abis'
    )
    if read_cache:
        from ctc import db

        abi = await db.async_query_contract_abi(
            address=contract_address,
            context=context,
        )
        if abi is not None:
            return abi

    from ctc.protocols import etherscan_utils

    # load from block explorer
    abi = await etherscan_utils.async_get_contract_abi(
        contract_address,
        context=context,
        verbose=verbose,
    )

    # get proxy implementation
    if proxy_implementation is None:
        proxy_implementation = (
            await contract_utils.async_get_proxy_implementation(
                contract_address=contract_address,
                context=context,
                block=block,
            )
        )

    # get proxy abi
    includes_proxy = False
    if proxy_implementation is not None:
        proxy_abi = await etherscan_utils.async_get_contract_abi(
            contract_address=proxy_implementation,
            context=context,
            verbose=verbose,
        )
        abi = contract_abi_modification.combine_contract_abis([abi, proxy_abi])
        includes_proxy = True

    # save to db
    if write_cache:
        from ctc import db

        await db.async_intake_contract_abi(
            contract_address=contract_address,
            context=context,
            abi=abi,
            includes_proxy=includes_proxy,
        )

    return abi




from __future__ import annotations

# see https://ethereum.github.io/beacon-APIs

consensus_endpoints = {
    #
    # beacon
    'beacon_genesis': '/eth/v1/beacon/genesis',
    'beacon_state_root': '/eth/v1/beacon/states/{state_id}/root',
    'beacon_state_fork': '/eth/v1/beacon/states/{state_id}/fork',
    'beacon_finality_checkpoints': '/eth/v1/beacon/states/{state_id}/finality_checkpoints',
    'beacon_validators': '/eth/v1/beacon/states/{state_id}/validators',
    'beacon_validator': '/eth/v1/beacon/states/{state_id}/validators/{validator_id}',
    'beacon_validator_balances': '/eth/v1/beacon/states/{state_id}/validator_balances',
    'beacon_committees': '/eth/v1/beacon/states/{state_id}/committees',
    'beacon_state_sync_committees': '/eth/v1/beacon/states/{state_id}/sync_committees',
    'beacon_blocks_headers': '/eth/v1/beacon/headers',
    'beacon_block_headers': '/eth/v1/beacon/headers/{block_id}',
    'beacon_blocks': '/eth/v1/beacon/blocks',
    'beacon_blinded_blocks': '/eth/v1/beacon/blinded_blocks',
    'beacon_block': '/eth/v2/beacon/blocks/{block_id}',
    'beacon_block_root': '/eth/v1/beacon/blocks/{block_id}/root',
    'beacon_block_attestations': '/eth/v1/beacon/blocks/{block_id}/attestations',
    'beacon_pool_attestations': '/eth/v1/beacon/pool/attestations',
    'beacon_attester_slashings': '/eth/v1/beacon/pool/attester_slashings',
    'beacon_proposer_slashings': '/eth/v1/beacon/pool/proposer_slashings',
    'beacon_pool_sync_committees': '/eth/v1/beacon/pool/sync_committees',
    'beacon_voluntary_exits': '/eth/v1/beacon/pool/voluntary_exits',
    #
    # config
    'config_fork_schedule': '/eth/v1/config/fork_schedule',
    'config_spec': '/eth/v1/config/spec',
    'config_deposit_contract': '/eth/v1/config/deposit_contract',
    #
    # node
    'node_identity': '/eth/v1/node/identity',
    'node_peers': '/eth/v1/node/peers',
    'node_peer': '/eth/v1/node/peers/{peer_id}',
    'node_peer_count': '/eth/v1/node/peer_count',
    'node_version': '/eth/v1/node/version',
    'node_syncing': '/eth/v1/node/syncing',
    'node_health': '/eth/v1/node/health',
}




from __future__ import annotations

import typing

if typing.TYPE_CHECKING:
    import tooltime

from ctc import spec
from .. import event_utils
from . import erc4626_normalize
from . import erc4626_spec


async def async_get_erc4626_deposits(
    token: spec.Address,
    *,
    start_block: typing.Optional[spec.BlockNumberReference] = None,
    end_block: typing.Optional[spec.BlockNumberReference] = None,
    start_time: tooltime.Timestamp | None = None,
    end_time: tooltime.Timestamp | None = None,
    include_timestamps: bool = False,
    verbose: bool = False,
    normalize: bool = True,
    convert_from_str: bool = True,
    context: spec.Context = None,
    **event_kwargs: typing.Any,
) -> spec.DataFrame:
    """get Deposit events for ERC-4626 vault"""

    import polars as pl

    event_abi = erc4626_spec.erc4626_event_abis['Deposit']

    deposits = await event_utils.async_get_events(
        contract_address=token,
        event_abi=event_abi,
        start_block=start_block,
        end_block=end_block,
        start_time=start_time,
        end_time=end_time,
        include_timestamps=include_timestamps,
        verbose=verbose,
        context=context,
        **event_kwargs,
    )

    if convert_from_str:
        deposits = deposits.with_columns(
            pl.col('arg__assets').apply(int),
            pl.col('arg__shares').apply(int),
        )

    if normalize and len(deposits) > 0:
        if not convert_from_str:
            raise Exception('must convert from str with convert_from_str=True')
        last_block = deposits['block_number'][-1]
        assets = await erc4626_normalize.async_normalize_erc4626_assets(
            token=token,
            assets=deposits['arg__assets'].to_list(),
            block=last_block,
            context=context,
        )
        deposits = deposits.with_columns(pl.Series('arg__assets', assets))

        shares = await erc4626_normalize.async_normalize_erc4626_shares(
            token=token,
            shares=deposits['arg__shares'].to_list(),
            block=last_block,
            context=context,
        )
        deposits = deposits.with_columns(pl.Series('arg__shares', shares))

    return deposits


async def async_get_erc4626_withdraws(
    token: spec.Address,
    *,
    start_block: typing.Optional[spec.BlockNumberReference] = None,
    end_block: typing.Optional[spec.BlockNumberReference] = None,
    start_time: tooltime.Timestamp | None = None,
    end_time: tooltime.Timestamp | None = None,
    include_timestamps: bool = False,
    verbose: bool = False,
    context: spec.Context = None,
    normalize: bool = True,
    convert_from_str: bool = True,
    **event_kwargs: typing.Any,
) -> spec.DataFrame:
    """get Withdraw events for ERC-4626 vault"""

    import polars as pl

    event_abi = erc4626_spec.erc4626_event_abis['Withdraw']

    withdraws = await event_utils.async_get_events(
        contract_address=token,
        event_abi=event_abi,
        start_block=start_block,
        end_block=end_block,
        start_time=start_time,
        end_time=end_time,
        include_timestamps=include_timestamps,
        verbose=verbose,
        context=context,
        **event_kwargs,
    )

    if convert_from_str:
        withdraws = withdraws.with_columns(
            pl.col('arg__assets').apply(int),
            pl.col('arg__shares').apply(int),
        )

    if normalize and len(withdraws) > 0:
        if not convert_from_str:
            raise Exception('must convert from str with convert_from_str=True')
        blocks = withdraws['block_number'].to_list()

        arg__assets = await erc4626_normalize.async_normalize_erc4626_assets(
            token=token,
            assets=withdraws['arg__assets'].to_list(),
            block=blocks[-1],
            context=context,
        )
        arg__shares = await erc4626_normalize.async_normalize_erc4626_shares(
            token=token,
            shares=withdraws['arg__shares'].to_list(),
            block=blocks[-1],
            context=context,
        )
        withdraws = withdraws.with_columns(
            pl.Series('arg__assets', arg__assets),
            pl.Series('arg__shares', arg__shares),
        )

    return withdraws



from __future__ import annotations

import typing

import toolsql

import ctc
from ctc import config
from ctc import evm
from ctc import spec

from .. import schema_utils


def is_schema_versions_initialized(conn: toolsql.Connection) -> bool:
    return 'schema_versions' in toolsql.get_table_names(conn)


def get_schema_version(
    schema_name: str,
    *,
    context: spec.Context,
    conn: toolsql.Connection,
) -> str | None:

    # use chain_id = -1 for dbs that have no network association
    if schema_name not in schema_utils.get_network_schema_names():
        chain_id = -1
    else:
        chain_id = config.get_context_chain_id(context)

    result = toolsql.select(
        conn=conn,
        table='schema_versions',
        where_equals={'chain_id': chain_id, 'schema_name': schema_name},
        columns=['version'],
        output_format='cell_or_none',
    )

    if result is not None and not isinstance(result, str):
        raise Exception('invalid result format')

    return result


def set_schema_version(
    schema_name: str,
    *,
    context: spec.Context,
    conn: toolsql.Connection,
    version: str | None = None,
) -> None:

    if version is None:
        version = ctc.__version__

    # use chain_id = -1 for dbs that have no network association
    if schema_name not in schema_utils.get_network_schema_names():
        chain_id = -1
    else:
        chain_id = config.get_context_chain_id(context)

    row = {'chain_id': chain_id, 'schema_name': schema_name, 'version': version}
    toolsql.insert(row=row, conn=conn, table='schema_versions')


def delete_schema_version(
    schema_name: str | None,
    *,
    network: spec.NetworkReference | None,
    conn: toolsql.Connection,
    confirm_delete_row: bool = False,
    confirm_delete_network: bool = False,
    confirm_delete_schema: bool = False,
    confirm_delete_all: bool = False,
) -> None:

    if schema_name is None and network is None:
        if not confirm_delete_all:
            import toolcli

            answer = toolcli.input_yes_or_no(
                'delete ALL version rows for ALL schemas? '
            )
            if not answer:
                raise Exception('aborting')

    elif schema_name is not None and network is not None:
        if not confirm_delete_row:
            import toolcli

            answer = toolcli.input_yes_or_no(
                'delete version row for schema '
                + schema_name
                + ' for network '
                + str(network)
                + ' ? '
            )
            if not answer:
                raise Exception('aborting')

    elif schema_name is not None:
        if not confirm_delete_schema:
            import toolcli

            answer = toolcli.input_yes_or_no(
                'delete ALL version rows for schema ' + schema_name + '? '
            )
            if not answer:
                raise Exception('aborting')

    elif network is not None:
        if not confirm_delete_network:
            import toolcli

            answer = toolcli.input_yes_or_no(
                'delete ALL version rows for network ' + str(network) + '? '
            )
            if not answer:
                raise Exception('aborting')

    where_equals: typing.MutableMapping[str, typing.Any] = {}
    if schema_name is not None:
        where_equals['schema_name'] = schema_name
    if network is not None:
        where_equals['chain_id'] = evm.get_network_chain_id(network)

    toolsql.delete(
        conn=conn,
        table='schema_versions',
        where_equals=where_equals,
    )




from __future__ import annotations

import typing

if typing.TYPE_CHECKING:
    import tooltime

from ctc import evm
from ctc import spec


async def async_predict_block_timestamp(
    block: typing.SupportsInt,
    *,
    context: spec.Context = None,
    window_size: int = 88295,
) -> int:
    """predict timestamp of future block number"""

    block = int(block)

    latest_block = await evm.async_get_block('latest', context=context)
    latest_number = latest_block['number']
    if block == latest_number:
        return latest_block['timestamp']
    elif block < latest_number:
        return await evm.async_get_block_timestamp(block, context=context)
    else:
        old_block = await evm.async_get_block(
            latest_number - window_size, context=context
        )
        mean_block_time = (
            latest_block['timestamp'] - old_block['timestamp']
        ) / (latest_number - old_block['number'])
        return round(
            latest_block['timestamp']
            + (block - latest_number) * mean_block_time
        )


async def async_predict_block_timestamps(
    blocks: typing.Sequence[typing.SupportsInt],
    *,
    context: spec.Context = None,
    window_size: int = 100000,
) -> typing.Sequence[int]:
    """predict timestamps of future block numbers"""

    import asyncio

    int_blocks = [int(block) for block in blocks]

    predictions = {}
    old_block_tasks = {}
    window_task = None
    new_blocks = []

    latest_block = await evm.async_get_block('latest', context=context)
    latest_number = latest_block['number']

    for block in int_blocks:
        if block == latest_number:
            predictions[block] = latest_block['timestamp']
        elif block < latest_number:
            coroutine = evm.async_get_block_timestamp(block, context=context)
            old_block_tasks[block] = asyncio.create_task(coroutine)
        else:
            if window_task is None:
                window_task = evm.async_get_block(
                    latest_number - window_size, context=context
                )
            new_blocks.append(block)

    # process new blocks
    if window_task is not None:
        old_block = await window_task
        mean_block_time = (
            latest_block['timestamp'] - old_block['timestamp']
        ) / (latest_number - old_block['number'])
        for new_block in new_blocks:
            new_timestamp = round(
                latest_block['timestamp']
                + (new_block - latest_number) * mean_block_time
            )
            predictions[new_block] = new_timestamp

    # gather old block tasks
    if len(old_block_tasks) > 0:
        old_block_timestamps = await asyncio.gather(*old_block_tasks.values())
        for block, timestamp in zip(
            old_block_tasks.keys(), old_block_timestamps
        ):
            predictions[block] = timestamp

    return [predictions[block] for block in int_blocks]


async def async_predict_timestamp_block(
    timestamp: tooltime.Timestamp,
    *,
    context: spec.Context = None,
    window_size: int = 86400 * 16,
) -> int:
    """predict block number of future timestamp"""

    import tooltime

    timestamp = tooltime.timestamp_to_seconds(timestamp)

    latest_block = await evm.async_get_block('latest', context=context)
    latest_timestamp = latest_block['timestamp']

    if timestamp == latest_timestamp:
        return latest_block['number']
    elif timestamp < latest_timestamp:
        return await evm.async_get_block_of_timestamp(
            timestamp, context=context
        )
    else:
        old_timestamp = latest_timestamp - window_size
        old_block = await evm.async_get_block_of_timestamp(
            old_timestamp, context=context
        )
        mean_blocks_per_time = (latest_block['number'] - old_block) / (
            latest_block['timestamp'] - old_timestamp
        )
        return round(
            latest_block['number']
            + (timestamp - latest_timestamp) * mean_blocks_per_time
        )


async def async_predict_timestamp_blocks(
    timestamps: typing.Sequence[tooltime.Timestamp],
    *,
    context: spec.Context = None,
    window_size: int = 86400 * 16,
) -> typing.Sequence[int]:
    """predict timestamps of future block numbers"""

    import asyncio
    import tooltime

    int_timestamps = [
        tooltime.timestamp_to_seconds(timestamp) for timestamp in timestamps
    ]

    latest_block = await evm.async_get_block('latest', context=context)
    latest_timestamp = latest_block['timestamp']
    old_timestamp_tasks = {}
    new_timestamps = []
    window_task = None

    predictions: dict[int, int | float] = {}
    for timestamp in int_timestamps:
        if timestamp == latest_timestamp:
            predictions[timestamp] = latest_block['number']
        elif timestamp < latest_timestamp:
            coroutine = evm.async_get_block_of_timestamp(
                timestamp, context=context
            )
            old_timestamp_tasks[timestamp] = asyncio.create_task(coroutine)
        else:
            if window_task is None:
                window_timestamp = latest_timestamp - window_size
                window_coroutine = evm.async_get_block_of_timestamp(
                    window_timestamp, context=context
                )
                window_task = asyncio.create_task(window_coroutine)
            new_timestamps.append(timestamp)

    if window_task is not None:
        window_block = await window_task
        mean_blocks_per_time = (latest_block['number'] - window_block) / (
            latest_timestamp - window_timestamp
        )
        for timestamp in new_timestamps:
            predictions[timestamp] = latest_block[
                'number'
            ] + mean_blocks_per_time * (timestamp - latest_timestamp)

    if len(old_timestamp_tasks) > 0:
        results = await asyncio.gather(*old_timestamp_tasks.values())
        for timestamp, result in zip(old_timestamp_tasks.keys(), results):
            predictions[timestamp] = result

    return [round(predictions[timestamp]) for timestamp in int_timestamps]




"""
# balances of wallets within single block
ctc eth balances <wallets> [--block <block>]

# balance of single wallet across multiple blocks
ctc eth balances <wallet> [--blocks <block>]
"""

from __future__ import annotations

import typing

import toolcli
import toolstr

import ctc.config
from ctc import cli
from ctc import evm
from ctc import spec
from ctc.cli import cli_utils


def get_command_spec() -> toolcli.CommandSpec:
    return {
        'f': async_balances_command,
        'help': 'output ETH balance across blocks or addresses',
        'args': [
            {
                'name': 'wallets',
                'nargs': '+',
                'help': 'wallet addresses to get balances of',
            },
            {'name': '--block', 'help': 'block number'},
            {
                'name': '--blocks',
                'nargs': '+',
                'help': 'block numbers to get balances in',
            },
            #
            {
                'name': '--raw',
                'action': 'store_true',
                'help': 'skip normalizing by 1e18 decimals',
            },
            {
                'name': '--export',
                'default': 'stdout',
                'help': 'file path for output (.json or .csv)',
            },
            {
                'name': '--overwrite',
                'action': 'store_true',
                'help': 'specify that output path can be overwritten',
            },
            {
                'name': ['-v', '--verbose'],
                'action': 'store_true',
                'help': 'display additional information',
            },
        ],
        'examples': {
            '0xd8da6bf26964af9d7eed9e03e53415d37aa96045 0xea674fdde714fd979de3edf0f56aa9716b898ec8 0xbe0eb53f46cd790cd13851d5eff43d12404d33e8': 'multiple wallets',
            '0xd8da6bf26964af9d7eed9e03e53415d37aa96045 --blocks 14000000:14000100': 'multiple blocks',
            '0xd8da6bf26964af9d7eed9e03e53415d37aa96045 -v --blocks 14000000:14100000:1000': 'multiple blocks with USD values',
        },
    }


async def async_balances_command(
    *,
    wallets: typing.Sequence[str],
    block: typing.Optional[spec.BlockNumberReference],
    blocks: typing.Optional[typing.Sequence[str]],
    raw: bool,
    export: str,
    overwrite: bool,
    verbose: bool,
) -> None:

    # REMOVE this
    import polars as pl

    indent = None
    wallets = [wallet.lower() for wallet in wallets]

    if blocks is not None:
        # single wallet, multiple blocks
        if len(wallets) > 1:
            raise Exception(
                'cannot specify multiple wallets and multiple blocks'
            )

        resolved_blocks = await cli_utils.async_parse_block_slice(blocks)
        if verbose:
            import asyncio
            from ctc.protocols import chainlink_utils

            eth_usd_coroutine = chainlink_utils.async_get_feed_data(
                'ETH_USD',
                blocks=resolved_blocks,
            )
            eth_usd_task = asyncio.create_task(eth_usd_coroutine)

        wallet = wallets[0]
        wallet = await evm.async_resolve_address(wallet, block=blocks[-1])
        balances = await evm.async_get_eth_balance_by_block(
            address=wallet,
            blocks=resolved_blocks,
            normalize=(not raw),
        )

        styles = cli.get_cli_styles()

        toolstr.print_text_box(
            'ETH Balances for ' + wallet, style=styles['title']
        )
        print()

        if verbose:
            eth_usd = (await eth_usd_task).to_numpy()
            usd_balances = [
                balance * eth_price
                for balance, eth_price in zip(balances, eth_usd[:, 0])
            ]
        rows = []
        for b in range(len(resolved_blocks)):
            row = [
                resolved_blocks[b],
                balances[b],
            ]
            if verbose:
                row.append(usd_balances[b])
                row.append(eth_usd[b])
            rows.append(row)

        labels = ['block', 'balance']
        if verbose:
            labels.append('$ balance')
            labels.append('ETH price')

        toolstr.print_table(
            rows=rows,
            labels=labels,
            border=styles['comment'],
            label_style=styles['title'],
            column_styles={
                'block': styles['metavar'],
                'balance': styles['description'],
                '$ balance': styles['description'],
                'ETH price': styles['description'],
            },
            column_formats={
                '$ balance': {
                    'prefix': '$',
                    'trailing_zeros': True,
                    'order_of_magnitude': True,
                },
                'ETH price': {'prefix': '$', 'trailing_zeros': True},
            },
            indent=4,
        )

        def formatter(xval: typing.Any) -> str:
            return toolstr.format(round(xval))

        xvals = resolved_blocks
        yvals = balances
        plot = toolstr.render_line_plot(
            xvals=xvals,
            yvals=yvals,
            n_rows=10,
            n_columns=60,
            line_style=styles['description'],
            chrome_style=styles['comment'],
            tick_label_style=styles['metavar'],
            xaxis_kwargs={'formatter': formatter},
            char_dict=ctc.config.get_cli_chart_charset(),
        )
        print()
        print()
        toolstr.print(
            toolstr.hjustify('ETH balance over time', 'center', 70),
            indent=4,
            style=styles['title'],
        )
        toolstr.print(plot, indent=4)

        if verbose:
            xvals = resolved_blocks
            yvals = usd_balances
            plot = toolstr.render_line_plot(
                xvals=xvals,
                yvals=yvals,
                n_rows=10,
                n_columns=60,
                line_style=styles['description'],
                chrome_style=styles['comment'],
                tick_label_style=styles['metavar'],
                xaxis_kwargs={'formatter': formatter},
                yaxis_kwargs={'tick_label_format': {'prefix': '$'}},
                char_dict=ctc.config.get_cli_chart_charset(),
            )
            print()
            print()
            toolstr.print(
                toolstr.hjustify('ETH balance over time (USD)', 'center', 70),
                indent=4,
                style=styles['title'],
            )
            toolstr.print(plot, indent=4)

        if export != 'stdout':
            df = pl.DataFrame(balances, ['balance'])
            df = df.with_columns(pl.Series('block', resolved_blocks))
            output_data: typing.Union[spec.DataFrame, spec.Series] = df
            cli_utils.output_data(
                output_data, export, overwrite=overwrite, indent=indent, raw=raw
            )
    else:
        # multiple wallets, single block
        if blocks is not None:
            raise Exception(
                'cannot specify multiple wallets and multiple blocks'
            )

        if block is None:
            block = 'latest'
        block = await evm.async_block_number_to_int(block)

        if wallets is not None:
            wallets = await evm.async_resolve_addresses(wallets, block=block)
            balances = await evm.async_get_eth_balance_of_addresses(
                addresses=wallets,
                block=block,
                normalize=(not raw),
            )

            output_data = pl.DataFrame(
                {'balance': balances, 'address': wallets}
            )

        cli_utils.output_data(
            output_data, export, overwrite=overwrite, indent=indent, raw=raw
        )



from __future__ import annotations

import typing

import toolsql

from ctc import spec
from ... import schema_utils


def _prepare_block_for_db(
    block: spec.DBBlock,
) -> typing.Mapping[str, typing.Any]:

    # remove extra keys
    extra_keys = ['send_root', 'l1_block_number', 'send_count', 'transactions']
    if any(key in block for key in extra_keys):
        block = block.copy()
        for key in extra_keys:
            if key in block:
                del block[key]  # type: ignore
    return block


async def async_upsert_block(
    *,
    block: spec.DBBlock,
    conn: toolsql.AsyncConnection,
    context: spec.Context,
) -> None:

    table = schema_utils.get_table_schema('blocks', context=context)
    ready_block = _prepare_block_for_db(block)
    await toolsql.async_insert(
        conn=conn,
        table=table,
        row=ready_block,
        upsert=True,
    )


async def async_upsert_blocks(
    *,
    blocks: typing.Sequence[spec.DBBlock],
    conn: toolsql.AsyncConnection,
    context: spec.Context,
) -> None:

    table = schema_utils.get_table_schema('blocks', context=context)
    ready_blocks = [_prepare_block_for_db(block) for block in blocks]
    await toolsql.async_insert(
        conn=conn,
        table=table,
        rows=ready_blocks,
        upsert=True,
    )


async def async_select_block(
    block_number: int | str,
    *,
    conn: toolsql.AsyncConnection,
    context: spec.Context,
) -> spec.DBBlock | None:

    table = schema_utils.get_table_schema('blocks', context=context)

    block: spec.DBBlock | None = await toolsql.async_select(  # type: ignore
        conn=conn,
        table=table,
        where_equals={'number': block_number},
        output_format='single_dict_or_none',
    )

    return block


async def async_select_blocks(
    block_numbers: typing.Sequence[int | str] | None = None,
    *,
    start_block: int | None = None,
    end_block: int | None = None,
    conn: toolsql.AsyncConnection,
    context: spec.Context,
) -> typing.Sequence[spec.DBBlock | None] | None:

    table = schema_utils.get_table_schema('blocks', context=context)

    if block_numbers is not None:
        blocks = await toolsql.async_select(
            conn=conn,
            table=table,
            where_in={'number': block_numbers},
        )

    elif start_block is not None and end_block is not None:
        blocks = await toolsql.async_select(
            conn=conn,
            table=table,
            where_gte={'number': start_block},
            where_lte={'number': end_block},
        )
        block_numbers = range(start_block, end_block + 1)

    else:
        raise Exception(
            'must specify block_numbers or start_block and end_block'
        )

    if blocks is None:
        return None

    for block in blocks:
        if block is not None and block['base_fee_per_gas'] is not None:
            block['base_fee_per_gas'] = int(block['base_fee_per_gas'])

    blocks_by_number = {
        block['number']: block for block in blocks if block is not None
    }

    return [blocks_by_number.get(number) for number in block_numbers]  # type: ignore


async def async_delete_block(
    block_number: int | str,
    *,
    conn: toolsql.AsyncConnection,
    context: spec.Context,
) -> None:

    table = schema_utils.get_table_schema('blocks', context=context)

    await toolsql.async_delete(
        conn=conn,
        table=table,
        where_equals={'number': block_number},
    )


async def async_delete_blocks(
    block_numbers: typing.Sequence[int | str] | None = None,
    *,
    start_block: int | None = None,
    end_block: int | None = None,
    conn: toolsql.AsyncConnection,
    context: spec.Context,
) -> None:

    table = schema_utils.get_table_schema('blocks', context=context)

    if block_numbers is not None:
        await toolsql.async_delete(
            conn=conn,
            table=table,
            where_in={'number': block_numbers},
        )
    elif start_block is not None and end_block is not None:
        await toolsql.async_delete(
            conn=conn,
            table=table,
            where_gte={'number': start_block},
            where_lte={'number': end_block},
        )
    else:
        raise Exception(
            'must specify block_numbrs or start_block and end_block'
        )


#
# # do not export these functions
#


async def async_select_block_timestamp(
    block_number: int,
    *,
    conn: toolsql.AsyncConnection,
    context: spec.Context = None,
) -> int | None:

    table = schema_utils.get_table_schema('blocks', context=context)

    result = await toolsql.async_select(
        conn=conn,
        table=table,
        where_equals={'number': block_number},
        columns=['timestamp'],
        output_format='cell_or_none',
    )
    if result is not None and not isinstance(result, int):
        raise Exception('invalid db result')
    return result


async def async_select_block_timestamps(
    block_numbers: typing.Sequence[typing.SupportsInt],
    *,
    conn: toolsql.AsyncConnection,
    context: spec.Context = None,
) -> list[int | None] | None:

    table = schema_utils.get_table_schema('blocks', context=context)

    block_numbers_int = [int(item) for item in block_numbers]

    results = await toolsql.async_select(
        conn=conn,
        table=table,
        where_in={'number': block_numbers_int},
    )

    if results is None:
        return None

    block_timestamps = {
        row['number']: row['timestamp'] for row in results if row is not None
    }

    return [
        block_timestamps.get(block_number) for block_number in block_numbers
    ]


async def async_select_max_block_number(
    *,
    conn: toolsql.AsyncConnection,
    context: spec.Context = None,
) -> int | None:

    table = schema_utils.get_table_schema('blocks', context=context)
    result = await toolsql.async_select(
        conn=conn,
        table=table,
        columns=['max(number)'],
        output_format='cell',
    )
    if result is not None:
        output = result['max__block_number']
        if output is not None and not isinstance(output, int):
            raise Exception('invalid db result')
        return output
    else:
        return None


async def async_select_max_block_timestamp(
    *,
    conn: toolsql.AsyncConnection,
    context: spec.Context = None,
) -> int | None:

    table = schema_utils.get_table_schema('blocks', context=context)
    result = await toolsql.async_select(
        conn=conn,
        table=table,
        columns=['max(number)'],
        output_format='cell',
    )
    if result is None:
        return None
    else:
        max_timestamp = result['max__timestamp']
        if max_timestamp is not None and not isinstance(max_timestamp, int):
            raise Exception('invalid db output')
        return max_timestamp


__all__ = (
    'async_upsert_block',
    'async_upsert_blocks',
    'async_select_block',
    'async_select_blocks',
    'async_delete_block',
    'async_delete_blocks',
)




from __future__ import annotations

from ctc import spec

from .. import format_utils
from .. import hash_utils
from . import secp256k1_utils


def private_key_to_public_key(private_key: spec.Data) -> str:
    """convert private key to public key"""

    binary_private_key = format_utils.to_binary(private_key)
    as_tuple = secp256k1_utils.privtopub(binary_private_key)
    return public_key_tuple_to_hex(as_tuple)


def private_key_to_address(private_key: spec.Data) -> spec.Address:
    """convert private key to address"""
    public_key = private_key_to_public_key(private_key)
    return public_key_to_address(public_key)


def public_key_to_address(
    public_key: tuple[int, int] | spec.Data
) -> spec.Address:
    """convert public key to address"""

    as_hex = public_key_tuple_to_hex(public_key)
    hash = hash_utils.keccak(as_hex, 'binary')
    address = hash[-20:]
    return format_utils.to_hex(address)


def public_key_tuple_to_hex(public_key: tuple[int, int] | spec.Data) -> str:
    """convert public key tuple to key"""

    if isinstance(public_key, tuple):
        if len(public_key) == 2:
            x, y = public_key
            as_bytes = x.to_bytes(32, 'big') + y.to_bytes(32, 'big')
            return format_utils.to_hex(as_bytes)
        else:
            raise Exception('unknown pubilc key format')

    else:
        return format_utils.to_hex(public_key, n_bytes=64)




import pytest

from ctc import evm


@pytest.mark.parametrize(
    'test',
    [
        ['USDC', '0xa0b86991c6218b36c1d19d4a2e9eb0ce3606eb48'],
    ],
)
async def test_get_erc20_address(test):
    symbol, address = test
    actual_address = await evm.async_get_erc20_address(symbol)
    assert address == actual_address


@pytest.mark.parametrize(
    'test',
    [['0xa0b86991c6218b36c1d19d4a2e9eb0ce3606eb48', 'USDC']],
)
async def test_get_erc20_symbol(test):
    address, symbol = test
    actual_symbol = await evm.async_get_erc20_symbol(address)
    assert symbol == actual_symbol



from __future__ import annotations

import asyncio
import typing

import toolcli
import toolstr
import tooltime

from ctc import cli
from ctc.cli import cli_utils
from ctc import evm
from ctc import config
from ctc import spec
from ctc.defi import dex_utils


help_message = """show information about a pool"""


def get_command_spec() -> toolcli.CommandSpec:
    return {
        'f': async_dex_pool_command,
        'help': help_message,
        'args': [
            {
                'name': 'pool',
                'help': 'address of pool',
            },
            {
                'name': '--network',
                'help': 'network to query',
            },
            {
                'name': '--block',
                'help': 'block number',
            },
        ],
        'examples': [
            '0xc5be99a02c6857f9eac67bbce58df5572498f40c',
            '0xc5be99a02c6857f9eac67bbce58df5572498f40c --block 14000000',
        ],
    }


async def async_dex_pool_command(
    *,
    pool: spec.Address,
    block: str | None,
    network: str,
    context: spec.Context = None,
) -> None:
    from ctc import db

    if block is not None:
        block_number = await cli_utils.async_parse_block(block, context=context)
    else:
        block_number = None

    dex_pool = await db.async_query_dex_pool(address=pool, context=context)
    if dex_pool is None:
        print('pool not found, updating dex database')
        await dex_utils.async_update_all_dexes()
        dex_pool = await db.async_query_dex_pool(address=pool, context=context)
        if dex_pool is None:
            raise Exception('could not find pool data')

    dex = await dex_utils.async_get_dex_class(
        pool=dex_pool['address'], context=context,
    )

    # queue creation timestamp task
    creation_timestamp_coroutine = evm.async_get_block_timestamp(
        dex_pool['creation_block'], context=context
    )
    creation_timestamp_task = asyncio.create_task(creation_timestamp_coroutine)

    # queue asset symbols task
    assets: typing.Sequence[str] = [
        typing.cast(str, dex_pool.get(key))
        for key in ['asset0', 'asset1', 'asset2', 'asset3']
        if dex_pool.get(key) is not None
    ]
    asset_symbols_task = asyncio.create_task(
        evm.async_get_erc20s_symbols(assets, context=context)
    )

    # queue balances task
    balances_coroutine = dex.async_get_pool_balances(
        pool,
        block=block_number,
        context=context,
    )
    balances_task = asyncio.create_task(balances_coroutine)

    # await all tasks
    creation_timestamp = await creation_timestamp_task
    asset_symbols = await asset_symbols_task
    balances = await balances_task

    # print bullet
    styles = cli.get_cli_styles()
    toolstr.print_text_box('Pool Summary', style=styles['title'])
    cli.print_bullet(
        key='pool',
        value=toolstr.add_style(dex_pool['address'], styles['metavar']),
    )
    cli.print_bullet(
        key='factory',
        value=toolstr.add_style(dex_pool['factory'], styles['metavar']),
    )
    dex_name = dex.get_dex_name()
    cli.print_bullet(key='DEX', value=dex_name)
    cli.print_bullet(key='creation block', value=dex_pool['creation_block'])
    cli.print_bullet(
        key='creation timestamp',
        value=tooltime.timestamp_to_iso_pretty(creation_timestamp),
    )
    cli.print_bullet(key='fee', value=dex_pool['fee'])
    cli.print_bullet(key='assets', value='')

    rows = []
    for a in range(len(assets)):
        row = [assets[a], asset_symbols[a], balances[assets[a]]]
        rows.append(row)
    labels = [
        'address',
        'symbol',
        'balance',
    ]
    toolstr.print_table(
        rows,
        labels=labels,
        indent=4,
        border=styles['comment'],
        label_style=styles['title'],
        column_styles={
            'address': styles['metavar'],
            'symbol': styles['option'],
            'balance': styles['description'],
        },
        column_formats={
            'balance': {'trailing_zeros': True},
        },
    )

    if dex_name == 'UniswapV2':
        from ctc.protocols import uniswap_v2_utils
        from ctc.defi.dex_utils.amm_utils import cpmm

        print()
        toolstr.print_header('Pool State', style=styles['title'])
        tokens_metadata = await uniswap_v2_utils.async_get_pool_tokens_metadata(
            pool,
            context=context,
        )
        x_symbol = tokens_metadata['x_symbol']
        y_symbol = tokens_metadata['y_symbol']
        pool_state = await uniswap_v2_utils.async_get_pool_state(
            pool,
            block=block_number,
            context=context,
        )
        cpmm.print_pool_summary(
            x_name=x_symbol,
            y_name=y_symbol,
            **pool_state,
        )



from __future__ import annotations

import typing

import msgspec

if typing.TYPE_CHECKING:
    SlotsData = typing.MutableMapping[
        tuple[str, str], typing.MutableMapping[str, typing.Any]
    ]


# SlotChanges = msgspec.defstruct(
#     'SlotChanges',
#     [
#         ('+', typing.Optional, None),
#         ('*', typing.Optional, None),
#         ('-', typing.Optional, None),
#     ],
#     omit_defaults=True,
# )

SlotAddress = str
ChangeType = str
SlotChanges = typing.Mapping[
    ChangeType, typing.Union[str, typing.Mapping[str, str]]
]


class ContractStateDiff(msgspec.Struct):
    storage: typing.Mapping[SlotAddress, SlotChanges]


class TraceResult(msgspec.Struct, rename='camel'):
    state_diff: typing.Mapping[str, ContractStateDiff]


class RpcResult(msgspec.Struct):
    result: list[TraceResult]


decoder = msgspec.json.Decoder(RpcResult)


def decode_slot_stats(
    raw_responses: typing.Sequence[str],
    block_numbers: typing.Sequence[int],
) -> SlotsData:
    slots_data: SlotsData = {}

    for raw_response, block_number in zip(raw_responses, block_numbers):
        response = decoder.decode(raw_response)
        for trace_result in response.result:
            for contract_address, state_diff in trace_result.state_diff.items():
                for slot, slot_changes in state_diff.storage.items():
                    key = (contract_address, slot)

                    # initialize slot
                    if key not in slots_data:
                        _initialize_slot(
                            slots_data=slots_data,
                            key=key,
                            slot_changes=slot_changes,
                            block_number=block_number,
                        )
                    else:
                        _integrate_slot_changes(
                            slots_data=slots_data,
                            key=key,
                            slot_changes=slot_changes,
                            block_number=block_number,
                        )

    return slots_data


def _initialize_slot(
    *,
    slots_data: SlotsData,
    key: tuple[str, str],
    slot_changes: typing.Mapping[str, typing.Any],
    block_number: int,
) -> None:
    if '+' in slot_changes:
        slots_data[key] = {
            'first_nonzero_block': block_number,
            'last_updated_block': block_number,
            'last_zero_block': None,
            'n_tx_updates': 1,
            'value': slot_changes['+'],
        }
    elif '*' in slot_changes:
        slots_data[key] = {
            'first_nonzero_block': block_number,
            'last_updated_block': block_number,
            'last_zero_block': None,
            'n_tx_updates': 1,
            'value': slot_changes['*']['to'],
        }
    elif '-' in slot_changes:
        slots_data[key] = {
            'first_nonzero_block': None,
            'last_updated_block': block_number,
            'last_zero_block': block_number,
            'n_tx_updates': 1,
            'value': '0x0',
        }
        raise Exception()
    else:
        raise Exception('unknown slot changes')


def _integrate_slot_changes(
    *,
    slots_data: SlotsData,
    key: tuple[str, str],
    slot_changes: typing.Mapping[str, typing.Any],
    block_number: int,
) -> None:
    """this function should be called for each block in increasing order"""

    datum = slots_data[key]
    datum['n_tx_updates'] += 1

    if '+' in slot_changes:
        datum['last_updated_block'] = block_number
        datum['value'] = slot_changes['+']
    elif '*' in slot_changes:
        datum['last_updated_block'] = block_number
        datum['value'] = slot_changes['*']['to']
    elif '-' in slot_changes:
        datum['last_zero_block'] = block_number
        datum['last_updated_block'] = block_number
        datum['value'] = '0x0'
    else:
        raise Exception('unknown slot changes')


#
# # flatten slot stats
#


def flatten_slots_data(
    slots_data: typing.Mapping[tuple[str, str], typing.Mapping[str, typing.Any]]
) -> typing.Sequence[typing.Mapping[str, typing.Any]]:
    return [
        dict(slot_data, contract_address=contract_address, slot=slot)
        for (contract_address, slot), slot_data in slots_data.items()
    ]


#
# # combining multiple slot datas
#


# def combine_slots_data(
#     slots_data_list: typing.Sequence[SlotsData],
# ) -> SlotsData:
#     if len(slots_data_list):
#         return {}
#     else:
#         slots_data: SlotsData = dict(slots_data_list[0])
#         for other_slots_data in slots_data_list[1:]:
#             for slot_address, other_slot_data in other_slots_data.items():
#                 _combine_slot_data(
#                     slots_data=slots_data,
#                     slot_address=slot_address,
#                     other_slot_data=other_slot_data,
#                 )
#     return slots_data


# def _combine_slot_data(
#     slots_data: SlotsData,
#     slot_address: tuple[str, str],
#     other_slot_data: typing.MutableMapping[str, typing.Any],
# ) -> None:
#     if slot_address not in slots_data:
#         slots_data[slot_address] = other_slot_data
#     else:
#         slot_data = slots_data[slot_address]

#         slots_data['n_tx_updates'] += other_slot_data['n_tx_updates']

#         if other_slot_data['first_nonzero_block'] is not None:
#             if slot_data['first_nonzero_block'] is None:
#                 slot_data['first_nonzero_block'] = other_slot_data[
#                     'first_nonzero_block'
#                 ]
#             elif (
#                 other_slot_data['first_nonzero_block']
#                 < slot_data['first_nonzero_block']
#             ):
#                 slot_data['first_nonzero_block'] = other_slot_data[
#                     'first_nonzero_block'
#                 ]

#         if other_slot_data['last_nonzero_block'] is not None:
#             if slot_data['last_nonzero_block'] is None:
#                 slot_data['last_nonzero_block'] = other_slot_data[
#                     'last_nonzero_block'
#                 ]
#             elif (
#                 other_slot_data['last_nonzero_block']
#                 > slot_data['last_nonzero_block']
#             ):
#                 slot_data['last_nonzero_block'] = other_slot_data[
#                     'last_nonzero_block'
#                 ]

#         if slot_data['last_updated_block'] is None:
#             slot_data['last_updated_block'] = other_slot_data[
#                 'last_updated_block'
#             ]
#         elif (
#             other_slot_data['last_updated_block']
#             > slot_data['last_updated_block']
#         ):
#             slot_data['last_updated_block'] = other_slot_data[
#                 'last_updated_block'
#             ]




from __future__ import annotations

import typing

import toolsql

from ctc import config

from .. import fourbyte_spec
from . import fourbyte_statements


#
# # functions
#


async def async_intake_function_signature(
    function_signature: fourbyte_spec.PartialEntry,
) -> None:

    db_config = config.get_context_db_config(schema_name='4byte', context=None)
    async with toolsql.async_connect(db_config) as conn:
        await fourbyte_statements.async_upsert_function_signature(
            function_signature=function_signature,
            conn=conn,
        )


async def async_intake_function_signatures(
    function_signatures: typing.Sequence[fourbyte_spec.PartialEntry],
) -> None:

    if len(function_signatures) == 0:
        return

    db_config = config.get_context_db_config(schema_name='4byte', context=None)
    async with toolsql.async_connect(db_config) as conn:
        await fourbyte_statements.async_upsert_function_signatures(
            function_signatures=function_signatures,
            conn=conn,
        )


#
# # events
#


async def async_intake_event_signature(
    event_signature: fourbyte_spec.PartialEntry,
) -> None:
    db_config = config.get_context_db_config(schema_name='4byte', context=None)
    async with toolsql.async_connect(db_config) as conn:
        await fourbyte_statements.async_upsert_event_signature(
            event_signature=event_signature,
            conn=conn,
        )


async def async_intake_event_signatures(
    event_signatures: typing.Sequence[fourbyte_spec.PartialEntry],
) -> None:
    db_config = config.get_context_db_config(schema_name='4byte', context=None)
    async with toolsql.async_connect(db_config) as conn:
        await fourbyte_statements.async_upsert_event_signatures(
            event_signatures=event_signatures,
            conn=conn,
        )




from __future__ import annotations

import json
import typing

import toolsql

from ctc import spec
from ... import schema_utils


async def async_upsert_contract_abi(
    *,
    address: spec.Address,
    abi: spec.ContractABI,
    includes_proxy: bool,
    conn: toolsql.AsyncConnection,
    context: spec.Context = None,
) -> None:

    abi_text = json.dumps(abi)

    table = schema_utils.get_table_schema('contract_abis', context=context)
    await toolsql.async_insert(
        conn=conn,
        table=table,
        row={
            'address': address.lower(),
            'abi_text': abi_text,
            'includes_proxy': includes_proxy,
        },
        upsert=True,
    )


async def async_select_contract_abi(
    address: spec.Address,
    *,
    context: spec.Context = None,
    conn: toolsql.AsyncConnection,
) -> spec.ContractABI | None:

    table = schema_utils.get_table_schema(
        'contract_abis',
        context=context,
    )
    abi_text = await toolsql.async_select(
        conn=conn,
        table=table,
        where_equals={'address': address.lower()},
        columns=['abi_text'],
        output_format='cell_or_none',
    )
    if abi_text is not None:
        contract_abi: spec.ContractABI = json.loads(abi_text)
        return contract_abi
    else:
        return None


async def async_select_contract_abis(
    addresses: typing.Sequence[spec.Address] | None = None,
    *,
    context: spec.Context = None,
    conn: toolsql.AsyncConnection,
) -> typing.Mapping[spec.Address, spec.ContractABI] | None:

    table = schema_utils.get_table_schema(
        'contract_abis',
        context=context,
    )

    if addresses is not None:
        where_in = {'address': addresses}
    else:
        where_in = None
    results = await toolsql.async_select(
        conn=conn,
        table=table,
        where_in=where_in,
    )

    if results is None:
        return None

    return {
        result['address']: json.loads(result['abi_text']) for result in results
    }


async def async_delete_contract_abi(
    address: spec.Address,
    *,
    conn: toolsql.AsyncConnection,
    context: spec.Context = None,
) -> None:

    table = schema_utils.get_table_schema(
        'contract_abis',
        context=context,
    )
    await toolsql.async_delete(
        conn=conn,
        table=table,
        where_equals={'address': address.lower()},
    )




from __future__ import annotations

import typing

from ctc import spec
from .. import block_utils


async def async_trace_contract_creations(
    start_block: spec.BlockNumberReference,
    end_block: spec.BlockNumberReference,
    *,
    context: spec.Context = None,
) -> typing.Sequence[typing.Any]:
    """collect contract creation traces"""
    import asyncio
    import ctc.rpc
    from ctc.rpc.rpc_decoders import create_trace_decoder

    start_block, end_block = await block_utils.async_resolve_block_range(
        start_block=start_block,
        end_block=end_block,
        to_int=True,
        allow_none=False,
        context=context,
    )

    blocks = list(range(start_block, end_block + 1))

    coroutines = [
        ctc.rpc.async_trace_block(
            block,
            decode_response=False,
            snake_case_response=False,
            raw_output=True,
            context=context,
        )
        for block in blocks
    ]
    result = await asyncio.gather(*coroutines)

    create_traces: typing.Sequence[typing.Any] = [
        trace
        for block_number, subresult in zip(blocks, result)
        for trace in create_trace_decoder.decode_create_traces(
            subresult, block_number=block_number
        )
    ]

    return create_traces


async def async_trace_native_transfers(
    start_block: spec.BlockNumberReference,
    end_block: spec.BlockNumberReference,
    *,
    context: spec.Context = None,
) -> typing.Sequence[typing.Any]:
    """collect native transfer traces"""
    # extract block replays from server
    import asyncio
    import ctc.rpc
    from ctc.rpc.rpc_decoders import native_transfer_decoder

    start_block, end_block = await block_utils.async_resolve_block_range(
        start_block=start_block,
        end_block=end_block,
        to_int=True,
        allow_none=False,
        context=context,
    )

    block_numbers = list(range(start_block, end_block + 1))
    coroutines = [
        ctc.rpc.async_trace_replay_block_transactions(
            block,
            trace_type=['trace'],
            decode_response=False,
            snake_case_response=False,
            raw_output=True,
            context=context,
        )
        for block in block_numbers
    ]
    responses = await asyncio.gather(*coroutines)

    native_transfers = native_transfer_decoder.decode_native_transfers(
        responses=responses, block_numbers=block_numbers
    )

    return native_transfers


async def async_trace_slot_stats(
    start_block: spec.BlockNumberReference,
    end_block: spec.BlockNumberReference,
    *,
    context: spec.Context = None,
) -> spec.DataFrame:
    """collect slot stat traces"""
    # extract block replays from server
    import asyncio
    import polars as pl
    import ctc.rpc
    from ctc.rpc.rpc_decoders import slot_diff_decoder

    start_block, end_block = await block_utils.async_resolve_block_range(
        start_block=start_block,
        end_block=end_block,
        to_int=True,
        allow_none=False,
        context=context,
    )

    block_numbers = list(range(start_block, end_block + 1))
    coroutines = [
        ctc.rpc.async_trace_replay_block_transactions(
            block,
            trace_type=['stateDiff'],
            decode_response=False,
            snake_case_response=False,
            raw_output=True,
            context=context,
        )
        for block in block_numbers
    ]
    responses = await asyncio.gather(*coroutines)

    # parse into slots data
    slot_data = slot_diff_decoder.decode_slot_stats(
        raw_responses=responses,
        block_numbers=block_numbers,
    )
    flat_slot_data = slot_diff_decoder.flatten_slots_data(slot_data)

    # convert to dataframe
    columns = [
        ('contract_address', pl.datatypes.Utf8),
        ('slot', pl.datatypes.Utf8),
        ('value', pl.datatypes.Utf8),
        ('first_nonzero_block', pl.datatypes.Int32),
        ('last_zero_block', pl.datatypes.Int32),
        ('last_updated_block', pl.datatypes.Int32),
        ('n_tx_updates', pl.datatypes.Int32),
    ]
    df = pl.DataFrame(flat_slot_data, schema=columns)

    return df




import pytest

import ctc


example_context_config = {
    'db_configs': {
        'local_sqlite': {'dbms': 'sqlite', 'path': ''},
        'local_postgres': {'dbms': 'sqlite', 'path': ''},
    },
    'context_cache_rules': [
        {
            'backend': 'local_sqlite',
            'read': True,
            'write': True,
        },
    ],
}

context_tests = [
    #
    # single boolean global cache
    {
        'context': {'cache': True},
        # default
        'schema_caches': {
            'blocks': ('local_sqlite', True, True),
            'transactions': ('local_sqlite', True, True),
        },
    },
    {
        'context': {'cache': False},
        'schema_caches': {
            'blocks': ('local_sqlite', False, False),
            'transactions': ('local_sqlite', False, False),
        },
    },
    #
    # single boolean read or write cache
    {
        'context': {'cache': {'read': False}},
        'schema_caches': {
            'blocks': ('local_sqlite', False, True),
            'transactions': ('local_sqlite', False, True),
        },
    },
    {
        'context': {'cache': {'read': True}},
        'schema_caches': {
            'blocks': ('local_sqlite', True, True),
            'transactions': ('local_sqlite', True, True),
        },
    },
    {
        'context': {'cache': {'write': False}},
        'schema_caches': {
            'blocks': ('local_sqlite', True, False),
            'transactions': ('local_sqlite', True, False),
        },
    },
    {
        'context': {'cache': {'write': True}},
        'schema_caches': {
            'blocks': ('local_sqlite', True, True),
            'transactions': ('local_sqlite', True, True),
        },
    },
    #
    # double boolean read or write cache
    {
        'context': {'cache': {'read': True, 'write': True}},
        'schema_caches': {
            'blocks': ('local_sqlite', True, True),
            'transactions': ('local_sqlite', True, True),
        },
    },
    {
        'context': {'cache': {'read': False, 'write': False}},
        'schema_caches': {
            'blocks': ('local_sqlite', False, False),
            'transactions': ('local_sqlite', False, False),
        },
    },
    {
        'context': {'cache': {'read': True, 'write': False}},
        'schema_caches': {
            'blocks': ('local_sqlite', True, False),
            'transactions': ('local_sqlite', True, False),
        },
    },
    {
        'context': {'cache': {'read': False, 'write': True}},
        'schema_caches': {
            'blocks': ('local_sqlite', False, True),
            'transactions': ('local_sqlite', False, True),
        },
    },
    #
    # single name of backend
    {
        'context': {'cache': 'local_sqlite'},
        'schema_caches': {
            'blocks': ('local_sqlite', True, True),
            'transactions': ('local_sqlite', True, True),
        },
    },
    {
        'context': {'cache': 'local_postgres'},
        'schema_caches': {
            'blocks': ('local_postgres', True, True),
            'transactions': ('local_postgres', True, True),
        },
    },
    #
    # single name of cache with single boolean read or write
    {
        'context': {'cache': {'backend': 'local_sqlite', 'read': True}},
        'schema_caches': {
            'blocks': ('local_sqlite', True, True),
            'transactions': ('local_sqlite', True, True),
        },
    },
    {
        'context': {'cache': {'backend': 'local_sqlite', 'write': True}},
        'schema_caches': {
            'blocks': ('local_sqlite', True, True),
            'transactions': ('local_sqlite', True, True),
        },
    },
    {
        'context': {'cache': {'backend': 'local_sqlite', 'read': False}},
        'schema_caches': {
            'blocks': ('local_sqlite', False, True),
            'transactions': ('local_sqlite', False, True),
        },
    },
    {
        'context': {'cache': {'backend': 'local_sqlite', 'write': False}},
        'schema_caches': {
            'blocks': ('local_sqlite', True, False),
            'transactions': ('local_sqlite', True, False),
        },
    },
    {
        'context': {'cache': {'backend': 'local_postgres', 'read': True}},
        'schema_caches': {
            'blocks': ('local_postgres', True, True),
            'transactions': ('local_postgres', True, True),
        },
    },
    {
        'context': {'cache': {'backend': 'local_postgres', 'write': True}},
        'schema_caches': {
            'blocks': ('local_postgres', True, True),
            'transactions': ('local_postgres', True, True),
        },
    },
    {
        'context': {'cache': {'backend': 'local_postgres', 'read': False}},
        'schema_caches': {
            'blocks': ('local_postgres', False, True),
            'transactions': ('local_postgres', False, True),
        },
    },
    {
        'context': {'cache': {'backend': 'local_postgres', 'write': False}},
        'schema_caches': {
            'blocks': ('local_postgres', True, False),
            'transactions': ('local_postgres', True, False),
        },
    },
    #
    # single name of cache with double boolean read or write
    {
        'context': {
            'cache': {'backend': 'local_sqlite', 'read': False, 'write': False}
        },
        'schema_caches': {
            'blocks': ('local_sqlite', False, False),
            'transactions': ('local_sqlite', False, False),
        },
    },
    {
        'context': {
            'cache': {'backend': 'local_sqlite', 'read': True, 'write': False}
        },
        'schema_caches': {
            'blocks': ('local_sqlite', True, False),
            'transactions': ('local_sqlite', True, False),
        },
    },
    {
        'context': {
            'cache': {'backend': 'local_sqlite', 'read': False, 'write': True}
        },
        'schema_caches': {
            'blocks': ('local_sqlite', False, True),
            'transactions': ('local_sqlite', False, True),
        },
    },
    {
        'context': {
            'cache': {'backend': 'local_sqlite', 'read': True, 'write': True}
        },
        'schema_caches': {
            'blocks': ('local_sqlite', True, True),
            'transactions': ('local_sqlite', True, True),
        },
    },
    {
        'context': {
            'cache': {
                'backend': 'local_postgres',
                'read': False,
                'write': False,
            }
        },
        'schema_caches': {
            'blocks': ('local_postgres', False, False),
            'transactions': ('local_postgres', False, False),
        },
    },
    {
        'context': {
            'cache': {'backend': 'local_postgres', 'read': True, 'write': False}
        },
        'schema_caches': {
            'blocks': ('local_postgres', True, False),
            'transactions': ('local_postgres', True, False),
        },
    },
    {
        'context': {
            'cache': {'backend': 'local_postgres', 'read': False, 'write': True}
        },
        'schema_caches': {
            'blocks': ('local_postgres', False, True),
            'transactions': ('local_postgres', False, True),
        },
    },
    {
        'context': {
            'cache': {'backend': 'local_postgres', 'read': True, 'write': True}
        },
        'schema_caches': {
            'blocks': ('local_postgres', True, True),
            'transactions': ('local_postgres', True, True),
        },
    },
    #
    # nested cache single boolean
    {
        'context': {'cache': {'blocks': False}},
        'schema_caches': {
            'blocks': ('local_sqlite', False, False),
            'transactions': ('local_sqlite', True, True),
        },
    },
    {
        'context': {'cache': {'blocks': True}},
        'schema_caches': {
            'blocks': ('local_sqlite', True, True),
            'transactions': ('local_sqlite', True, True),
        },
    },
    {
        'context': {'cache': {'transactions': False}},
        'schema_caches': {
            'blocks': ('local_sqlite', True, True),
            'transactions': ('local_sqlite', False, False),
        },
    },
    {
        'context': {'cache': {'transactions': True}},
        'schema_caches': {
            'blocks': ('local_sqlite', True, True),
            'transactions': ('local_sqlite', True, True),
        },
    },
    #
    # nested cache two boolean
    {
        'context': {'cache': {'blocks': False, 'transactions': False}},
        'schema_caches': {
            'blocks': ('local_sqlite', False, False),
            'transactions': ('local_sqlite', False, False),
        },
    },
    {
        'context': {'cache': {'blocks': True, 'transactions': False}},
        'schema_caches': {
            'blocks': ('local_sqlite', True, True),
            'transactions': ('local_sqlite', False, False),
        },
    },
    {
        'context': {'cache': {'blocks': False, 'transactions': True}},
        'schema_caches': {
            'blocks': ('local_sqlite', False, False),
            'transactions': ('local_sqlite', True, True),
        },
    },
    {
        'context': {'cache': {'blocks': True, 'transactions': True}},
        # default
        'schema_caches': {
            'blocks': ('local_sqlite', True, True),
            'transactions': ('local_sqlite', True, True),
        },
    },
    #
    # nested cache boolean mixed with name
    {
        'context': {'cache': {'blocks': 'local_sqlite', 'transactions': False}},
        'schema_caches': {
            'blocks': ('local_sqlite', True, True),
            'transactions': ('local_sqlite', False, False),
        },
    },
    {
        'context': {'cache': {'blocks': 'local_sqlite', 'transactions': True}},
        'schema_caches': {
            'blocks': ('local_sqlite', True, True),
            'transactions': ('local_sqlite', True, True),
        },
    },
    {
        'context': {
            'cache': {'blocks': 'local_postgres', 'transactions': False}
        },
        'schema_caches': {
            'blocks': ('local_postgres', True, True),
            'transactions': ('local_sqlite', False, False),
        },
    },
    {
        'context': {
            'cache': {'blocks': 'local_postgres', 'transactions': True}
        },
        'schema_caches': {
            'blocks': ('local_postgres', True, True),
            'transactions': ('local_sqlite', True, True),
        },
    },
    {
        'context': {'cache': {'blocks': False, 'transactions': 'local_sqlite'}},
        'schema_caches': {
            'blocks': ('local_sqlite', False, False),
            'transactions': ('local_sqlite', True, True),
        },
    },
    {
        'context': {'cache': {'blocks': True, 'transactions': 'local_sqlite'}},
        'schema_caches': {
            'blocks': ('local_sqlite', True, True),
            'transactions': ('local_sqlite', True, True),
        },
    },
    {
        'context': {
            'cache': {'blocks': False, 'transactions': 'local_postgres'}
        },
        'schema_caches': {
            'blocks': ('local_sqlite', False, False),
            'transactions': ('local_postgres', True, True),
        },
    },
    {
        'context': {
            'cache': {'blocks': True, 'transactions': 'local_postgres'}
        },
        'schema_caches': {
            'blocks': ('local_sqlite', True, True),
            'transactions': ('local_postgres', True, True),
        },
    },
    #
    # nested cache two names
    {
        'context': {
            'cache': {
                'blocks': 'local_sqlite',
                'transactions': 'local_sqlite',
            },
        },
        'schema_caches': {
            'blocks': ('local_sqlite', True, True),
            'transactions': ('local_sqlite', True, True),
        },
    },
    {
        'context': {
            'cache': {
                'blocks': 'local_postgres',
                'transactions': 'local_sqlite',
            },
        },
        'schema_caches': {
            'blocks': ('local_postgres', True, True),
            'transactions': ('local_sqlite', True, True),
        },
    },
    {
        'context': {
            'cache': {
                'blocks': 'local_sqlite',
                'transactions': 'local_postgres',
            },
        },
        'schema_caches': {
            'blocks': ('local_sqlite', True, True),
            'transactions': ('local_postgres', True, True),
        },
    },
    {
        'context': {
            'cache': {
                'blocks': 'local_postgres',
                'transactions': 'local_postgres',
            },
        },
        'schema_caches': {
            'blocks': ('local_postgres', True, True),
            'transactions': ('local_postgres', True, True),
        },
    },
    #
    # [double nested versions of everything above]
    # ...
]


# add double nested tests
default_cache_config = example_context_config['context_cache_rules'][-1]
default_tuple = (
    default_cache_config['backend'],
    default_cache_config['read'],
    default_cache_config['write'],
)
subconfigs = [
    {'backend': 'local_sqlite'},
    {'backend': 'local_postgres'},
    {'read': True},
    {'read': False},
    {'write': True},
    {'write': False},
    {'read': True, 'write': True},
    {'read': True, 'write': False},
    {'read': False, 'write': True},
    {'read': False, 'write': False},
    {'backend': 'local_sqlite', 'read': True},
    {'backend': 'local_sqlite', 'read': False},
    {'backend': 'local_sqlite', 'write': True},
    {'backend': 'local_sqlite', 'write': False},
    {'backend': 'local_sqlite', 'read': True, 'write': True},
    {'backend': 'local_sqlite', 'read': True, 'write': False},
    {'backend': 'local_sqlite', 'read': False, 'write': True},
    {'backend': 'local_sqlite', 'read': False, 'write': False},
    {'backend': 'local_postgres', 'read': True},
    {'backend': 'local_postgres', 'read': False},
    {'backend': 'local_postgres', 'write': True},
    {'backend': 'local_postgres', 'write': False},
    {'backend': 'local_postgres', 'read': True, 'write': True},
    {'backend': 'local_postgres', 'read': True, 'write': False},
    {'backend': 'local_postgres', 'read': False, 'write': True},
    {'backend': 'local_postgres', 'read': False, 'write': False},
]
for subconfig in subconfigs:

    full_subconfig = dict(default_cache_config, **subconfig)
    subconfig_tuple = (
        full_subconfig['backend'],
        full_subconfig['read'],
        full_subconfig['write'],
    )

    test_1 = {
        'context': {'cache': {'blocks': subconfig}},
        'schema_caches': {
            'blocks': subconfig_tuple,
            'transactions': default_tuple,
        },
    }
    test_2 = {
        'context': {'cache': {'transactions': subconfig}},
        'schema_caches': {
            'blocks': default_tuple,
            'transactions': subconfig_tuple,
        },
    }
    test_3 = {
        'context': {'cache': {'blocks': subconfig, 'transactions': subconfig}},
        'schema_caches': {
            'blocks': subconfig_tuple,
            'transactions': subconfig_tuple,
        },
    }
    test_4 = {
        'context': {'cache': {'blocks': subconfig, 'transactions': False}},
        'schema_caches': {
            'blocks': subconfig_tuple,
            'transactions': ('local_sqlite', False, False),
        },
    }
    test_5 = {
        'context': {'cache': {'blocks': False, 'transactions': subconfig}},
        'schema_caches': {
            'blocks': ('local_sqlite', False, False),
            'transactions': subconfig_tuple,
        },
    }

    context_tests.append(test_1)
    context_tests.append(test_2)
    context_tests.append(test_3)
    context_tests.append(test_4)
    context_tests.append(test_5)


@pytest.fixture()
def set_example_context_config():
    import ctc.config

    for key, value in example_context_config.items():
        ctc.config.set_config_override(key=key, value=value)

    yield

    ctc.config.clear_all_config_overrides()


@pytest.mark.parametrize('test', context_tests)
def test_get_context_schema_cache(test, set_example_context_config):
    context = test['context']

    schema_caches = test.get('schema_caches')
    if schema_caches is not None:
        for schema_name, target_schema_cache in schema_caches.items():
            actual_backend = ctc.config.get_context_cache_backend(
                schema_name=schema_name, context=context
            )
            read_cache, write_cache = ctc.config.get_context_cache_read_write(
                schema_name=schema_name, context=context
            )
            assert (
                actual_backend,
                read_cache,
                write_cache,
            ) == target_schema_cache


#
# # test interaction with config variables
#

# def test_global_cache_override():
#     pass


# def test_global_cache_override():
#     pass


# def test_global_cache_override():
#     pass


# def test_global_cache_override():
#     pass




from __future__ import annotations

import typing

import aiohttp
import toolcli
import toolstr
import toolsql

from ctc import cli
from ctc import db
from ctc import spec
from ... import config_defaults


def setup_dbs(
    *,
    styles: toolcli.StyleTheme,
    data_dir: str,
    network_data: spec.PartialConfig,
    headless: bool,
    old_db_configs: typing.Mapping[str, toolsql.DBConfig] | None = None,
) -> spec.PartialConfig:
    print()
    print()
    toolstr.print('## Database Setup', style=styles['title'])
    print()
    print('ctc stores its collected chain data in an sql database')

    # decide whether to use old config
    if old_db_configs is not None and old_db_configs.get('main') is not None:
        old_db_config = old_db_configs['main']
    else:
        old_db_config = None
    if old_db_config is not None:
        print()
        print('Current database config:')
        for key, value in old_db_config.items():
            if key == 'password':
                value = '********'
            cli.print_bullet(key=key, value=value)
        print()
        answer = toolcli.input_yes_or_no(
            prompt='Continue using this database configuration? ',
            default='yes',
            style=styles['metavar'],
            headless=headless,
        )
        if answer:
            db_configs = old_db_configs
        else:
            db_configs = None
    else:
        db_configs = None

    # initialize new config if old one not used
    if db_configs is None:
        print()
        dbms_choices = ['sqlite', 'postgresql']
        dbms_index = toolcli.input_number_choice(
            prompt='Which database do you want to use?\nuse sqlite for easy setup, use postgresql if in cloud',
            choices=dbms_choices,
            style=styles['metavar'],
            headless=headless,
            default='sqlite',
        )
        default_dbms = dbms_choices[dbms_index]
        if default_dbms == 'sqlite':
            db_configs = config_defaults.get_default_db_configs(
                data_dir=data_dir
            )
        else:
            db_configs = _get_postgres_config(headless=headless, styles=styles)

    # check database versions
    check_db_versions(list(db_configs.values()))

    # create db
    print()
    for db_config in db_configs.values():
        toolsql.create_db(
            db_config=db_config,
            db_schema=None,
            if_not_exists=True,
            verbose=True,
            confirm=True,
        )

    print()
    try:
        _delete_incomplete_chainlink_schemas(db_configs['main'])
    except toolsql.CannotConnect:
        pass

    # create tables
    used_networks: set[spec.NetworkReference] = set()
    default_network = network_data.get('default_network')
    if default_network is not None:
        used_networks.add(default_network)
    for provider in network_data['providers'].values():
        network = provider.get('network')
        if network is not None:
            used_networks.add(network)
    used_networks = {
        network for network in used_networks if network is not None
    }
    print()

    with toolsql.connect(db_configs['main']) as conn:
        db.create_missing_tables(
            networks=list(used_networks),
            conn=conn,
            confirm=True,
        )

    return {'db_configs': db_configs}


def _get_postgres_config(
    headless: bool,
    styles: toolcli.StyleTheme,
) -> typing.Mapping[str, toolsql.DBConfig]:
    print()
    print('Setting up PostgreSQL...')
    hostname = toolcli.input_prompt(
        'What is the hostname ',
        default='localhost',
        style=styles['metavar'],
        headless=headless,
    )
    port = toolcli.input_int(
        'What is the port? ',
        default='5432',
        style=styles['metavar'],
        headless=headless,
    )
    database = toolcli.input_prompt(
        'What is the database name ',
        default='ctc',
        style=styles['metavar'],
        headless=headless,
    )
    username = toolcli.input_prompt(
        'What is the username? ',
        default='postgres',
        style=styles['metavar'],
        headless=headless,
    )
    password = toolcli.input_prompt(
        'What is the password? ',
        default='',
        style=styles['metavar'],
        headless=headless,
    )
    db_config: toolsql.DBConfig = {
        'dbms': 'postgresql',
        'database': database,
        'hostname': hostname,
        'port': port,
        'username': username,
        'password': password,
    }

    return {'main': db_config}


async def async_populate_db_tables(
    db_config: toolsql.DBConfig,
    styles: toolcli.StyleTheme,
) -> None:
    from ctc.protocols.chainlink_utils import chainlink_db
    from ..default_data import default_erc20s

    print()
    print()
    toolstr.print('## Populating Database', style=styles['title'])

    # populate data: erc20s
    print()
    print('Populating database with metadata of common ERC20 tokens...')
    print()
    await default_erc20s.async_intake_default_erc20s(
        context=dict(network='ethereum'),
    )

    # populate data: chainlink
    print()
    print('Populating database with latest Chainlink oracle feeds...')
    print()
    try:
        await chainlink_db.async_import_networks_to_db(db_config=db_config)
    except aiohttp.client_exceptions.ClientConnectorError:
        print('Could not connect to Chainlink server, skipping')
    except Exception:
        print('Could not add feeds to db, skipping')


def _delete_incomplete_chainlink_schemas(db_config: toolsql.DBConfig) -> None:
    """detect any tables missing in chainlink schema

    this is a stopgap until a more comprehensive migration system is in place
    """

    from ctc import db

    # looking for schemas that have already been created, but are missing tables
    networks = list(config_defaults.get_default_networks_metadata().keys())
    with toolsql.connect(db_config) as conn:
        table_names = toolsql.get_table_names(conn)
        if 'schema_versions' not in table_names:
            return
        for network in networks:
            context: spec.Context = {'network': network}
            schema_version = db.get_schema_version(
                schema_name='chainlink',
                context=dict(network=network),
                conn=conn,
            )
            if schema_version is not None:
                schema = db.get_prepared_schema(
                    schema_name='chainlink', context=context
                )
                for table_name in schema['tables'].keys():
                    if table_name not in table_names:
                        print(
                            'missing chainlink_aggregator_updates table, rebuilding schema'
                        )
                        db.drop_schema(
                            schema_name='chainlink',
                            context=context,
                            confirm=True,
                            conn=conn,
                        )


def check_db_versions(db_configs: typing.Sequence[toolsql.DBConfig]) -> None:
    dbms_set = {db_config.get('dbms') for db_config in db_configs}
    for dbms in dbms_set:
        # check sqlite
        if dbms == 'sqlite':
            import sqlite3

            # get sqlite3 version
            if sqlite3.sqlite_version.count('.') == 2:
                major_str, minor_str, _ = sqlite3.sqlite_version.split('.')
            elif sqlite3.sqlite_version.count('.') == 1:
                major_str, minor_str = sqlite3.sqlite_version.split('.')
            else:
                major_str, minor_str = '0', '0'
            major = int(major_str)
            minor = int(minor_str)

            if (major < 3) or (major == 3 and minor < 24):
                raise Exception(
                    'ctc requires sqlite verison >= 3.24. This environment is using sqlite version '
                    + str(sqlite3.sqlite_version)
                    + '. You must upgrade sqlite3 before continuing.'
                    + ' If using apt, this can be accomplished using `apt install sqlite3` or `sudo apt-get install sqlite3`'
                )

        elif dbms == 'postgresql':
            pass

        else:
            raise Exception('dbms not supported: ' + str(dbms))




from __future__ import annotations

from .. import binary_utils
from ctc import spec


def serialize_block(block: spec.RPCBlock) -> spec.PrefixHexData:
    """serialize block in preparation for hashing"""

    if block.get('base_fee_per_gas') is not None:
        block_type = 'eip1559'
    else:
        block_type = 'legacy'

    as_list = [
        block['parent_hash'],
        block['sha3_uncles'],
        block['miner'],
        block['state_root'],
        block['transactions_root'],
        block['receipts_root'],
        block['logs_bloom'],
        block['difficulty'],
        block['number'],
        block['gas_limit'],
        block['gas_used'],
        block['timestamp'],
        block['extra_data'],
        block['mix_hash'],
        block['nonce'],
    ]

    if block_type == 'eip1559':
        as_list.append(block['base_fee_per_gas'])

    return binary_utils.rlp_encode(as_list)


def hash_block(block: spec.RPCBlock) -> spec.PrefixHexData:
    """compute hash of block"""

    serialized = serialize_block(block)
    return binary_utils.keccak(serialized)



from __future__ import annotations

import typing

from ctc import spec


#
# # basic conversion
#

def convert_rpc_transaction_to_db_transaction(
    transaction: spec.RPCTransaction,
    receipt: spec.RPCTransactionReceipt,
) -> spec.DBTransaction:
    """convert transaction data from rpc call into db transaction format"""

    if set(transaction.keys()) == spec.db_transaction_keys:
        return transaction  # type: ignore

    to = transaction.get('to')
    if to is None:
        to = '0x0000000000000000000000000000000000000000'

    tx: spec.DBTransaction = {
        #
        # tx fields
        'hash': transaction['hash'],
        'block_number': transaction['block_number'],
        'transaction_index': transaction['transaction_index'],
        'to_address': to,
        'from_address': transaction['from'],
        'value': transaction['value'],
        'input': transaction['input'],
        'nonce': transaction['nonce'],
        'transaction_type': transaction['type'],
        'access_list': transaction.get('access_list'),
        #
        # receipt fields
        'gas_used': receipt['gas_used'],
        'gas_price': receipt['effective_gas_price'],
        'gas_limit': transaction['gas'],
        'gas_priority': transaction.get('max_priority_fee_per_gas'),
        'gas_price_max': transaction.get('max_fee_per_gas'),
        'status': bool(receipt['status']),
    }
    return tx


#
# # conversions that require fetching data
#

async def async_convert_rpc_transaction_to_db_transaction(
    transaction: spec.DBTransaction | spec.RPCTransaction,
    *,
    context: spec.Context = None,
) -> spec.DBTransaction:
    """convert transaction to standard form, fetching receipts as necessary"""

    from ctc import rpc

    if set(transaction.keys()) == spec.db_transaction_keys:
        return transaction  # type: ignore
    else:
        receipt = await rpc.async_eth_get_transaction_receipt(
            transaction_hash=transaction['hash'],
            context=context,
        )
        return convert_rpc_transaction_to_db_transaction(
            transaction=transaction,  # type: ignore
            receipt=receipt,
        )


async def async_convert_rpc_transactions_to_db_transactions(
    transactions: typing.Sequence[spec.DBTransaction | spec.RPCTransaction],
    *,
    context: spec.Context = None,
) -> typing.Sequence[spec.DBTransaction]:
    """convert transactions to standard form, fetching receipts as necessary"""

    import asyncio

    coroutines = [
        async_convert_rpc_transaction_to_db_transaction(
            transaction=transaction,
            context=context,
        )
        for transaction in transactions
    ]
    return await asyncio.gather(*coroutines)


#
# # text <--> int conversion
#

def convert_db_transaction_fields_to_text(
    transaction: spec.DBTransaction,
    *,
    inplace: bool = False,
) -> spec.DBTransactionText:
    """convert large integer transaction fields from int to text"""
    if not inplace:
        transaction = transaction.copy()
    field: typing.Literal['value']
    for field in ['value']:  # type: ignore
        transaction[field] = str(transaction[field])  # type: ignore
    return transaction  # type: ignore


def convert_db_transaction_fields_to_int(
    transaction: spec.DBTransactionText,
    *,
    inplace: bool = False,
) -> spec.DBTransaction:
    """convert large integer transaction fields from text to int"""
    if not inplace:
        transaction = transaction.copy()
    field: typing.Literal['value']
    for field in ['value']:  # type: ignore
        transaction[field] = int(transaction[field])  # type: ignore
    return transaction  # type: ignore




from __future__ import annotations

import typing

from typing_extensions import TypedDict, NotRequired

from . import address_types
from . import binary_types


class EncodedEvent(TypedDict):
    block_number: int
    transaction_index: int
    log_index: int
    transaction_hash: str
    contract_address: address_types.Address
    event_hash: str
    topic1: typing.Any
    topic2: typing.Any
    topic3: typing.Any
    unindexed: typing.Any


class EventQuery(TypedDict):
    contract_address: address_types.Address | None
    event_hash: typing.Any | None
    topic1: binary_types.BinaryData | None
    topic2: binary_types.BinaryData | None
    topic3: binary_types.BinaryData | None
    start_block: int
    end_block: int


class DBEventQuery(TypedDict):
    query_id: NotRequired[int]
    query_type: int
    contract_address: address_types.Address | None
    event_hash: typing.Any | None
    topic1: binary_types.BinaryData | None
    topic2: binary_types.BinaryData | None
    topic3: binary_types.BinaryData | None
    start_block: int
    end_block: int


class EventQueryPlan(TypedDict):
    db: typing.Sequence[EventQuery]
    node: typing.Sequence[EventQuery]



from __future__ import annotations

import typing

from ctc import evm
from ctc import spec

from .. import dex_class
from . import uniswap_v2_dex

if typing.TYPE_CHECKING:
    import tooltime


class UniswapV3DEX(dex_class.DEX):
    """Uniswap V3 DEX"""

    _pool_factories = {1: ['0x1f98431c8ad98523631ae4a59f267346ea31f984']}

    @classmethod
    async def async_get_new_pools(
        cls,
        *,
        factory: spec.Address,
        start_block: spec.BlockNumberReference | None = None,
        end_block: spec.BlockNumberReference | None = None,
        start_time: tooltime.Timestamp | None = None,
        end_time: tooltime.Timestamp | None = None,
        context: spec.Context = None,
    ) -> typing.Sequence[spec.DexPool]:

        from ctc.protocols import uniswap_v3_utils

        df = await evm.async_get_events(
            factory,
            event_abi=uniswap_v3_utils.factory_event_abis['PoolCreated'],
            verbose=False,
            start_block=start_block,
            end_block=end_block,
            start_time=start_time,
            end_time=end_time,
            context=context,
        )

        dex_pools = []
        for row in df.to_dicts():
            dex_pool: spec.DexPool = {
                'address': row['arg__pool'],
                'factory': factory,
                'asset0': row['arg__token0'],
                'asset1': row['arg__token1'],
                'asset2': None,
                'asset3': None,
                'fee': row['arg__fee'] * 100,
                'creation_block': row['block_number'],
                'additional_data': {},
            }
            dex_pools.append(dex_pool)

        return dex_pools

    @classmethod
    async def _async_get_pool_assets_from_node(
        cls,
        pool: spec.Address,
        *,
        block: spec.BlockNumberReference | None = None,
        context: spec.Context = None,
    ) -> tuple[str, str]:
        output = uniswap_v2_dex.UniswapV2DEX._async_get_pool_assets_from_node(
            pool=pool,
            block=block,
            context=context,
        )
        return await output

    @classmethod
    async def _async_get_pool_raw_trades(
        cls,
        pool: spec.Address,
        *,
        start_block: spec.BlockNumberReference | None = None,
        end_block: spec.BlockNumberReference | None = None,
        start_time: tooltime.Timestamp | None = None,
        end_time: tooltime.Timestamp | None = None,
        include_timestamps: bool = False,
        verbose: bool = False,
        context: spec.Context = None,
    ) -> spec.RawDexTrades:

        from ctc.protocols import uniswap_v3_utils

        event_abi = await uniswap_v3_utils.async_get_event_abi('Swap', 'pool')

        trades = await evm.async_get_events(
            event_abi=event_abi,
            contract_address=pool,
            start_block=start_block,
            end_block=end_block,
            start_time=start_time,
            end_time=end_time,
            include_timestamps=include_timestamps,
            verbose=verbose,
            context=context,
        )

        bool_bought_id = trades['arg__amount0'].apply(int) > 0
        bought_id = bool_bought_id.apply(int)
        sold_id = (bought_id == 0).apply(int)
        sold_amount = sold_id * trades['arg__amount1'].apply(
            int
        ) + bought_id * trades['arg__amount0'].apply(int)
        bought_amount = -(
            bought_id * trades['arg__amount1'].apply(int)
            + sold_id * trades['arg__amount0'].apply(int)
        )

        output: spec.RawDexTrades = {
            'block_number': trades['block_number'],
            'transaction_hash': trades['transaction_hash'],
            'recipient': trades['arg__recipient'],
            'sold_id': sold_id,
            'bought_id': bought_id,
            'sold_amount': sold_amount,
            'bought_amount': bought_amount,
        }

        return output



import subprocess

import sys

import pytest
import toolcli

from ctc.cli import cli_run


@pytest.mark.parametrize('spec_reference', list(cli_run.command_index.values()))
def test_subcommands_have_spec(spec_reference):
    command_spec = toolcli.resolve_command_spec(spec_reference)
    assert isinstance(command_spec, dict), 'missing command_spec for ' + str(
        spec_reference
    )
    allowed_keys = set(toolcli.CommandSpec.__annotations__.keys())
    actual_keys = set(command_spec.keys())
    assert set(actual_keys).issubset(
        allowed_keys
    ), 'keys not allowed in ' + str(spec_reference)


@pytest.mark.parametrize('spec_reference', list(cli_run.command_index.values()))
def test_subcommands_have_help(spec_reference):
    command_spec = toolcli.resolve_command_spec(spec_reference)
    assert (
        command_spec.get('help') is not None
    ), 'missing help for command for ' + str(spec_reference)
    for arg_spec in command_spec.get('args', []):
        assert (
            arg_spec.get('help') is not None
        ), 'missing help for arg in ' + str(spec_reference)


@pytest.mark.parametrize('spec_reference', list(cli_run.command_index.values()))
def test_subcommands_have_examples(spec_reference):
    command_spec = toolcli.resolve_command_spec(spec_reference)
    assert (
        command_spec.get('examples') is not None
    ), 'missing examples for command for ' + str(spec_reference)


def collect_cli_examples():

    cli_examples = []
    for subcommand, spec_reference in cli_run.command_index.items():
        command_spec = toolcli.resolve_command_spec(spec_reference)
        if command_spec.get('hidden', False):
            continue

        # collect non-hidden examples
        example_strs = []
        examples = command_spec.get('examples', [])
        if isinstance(examples, list):
            example_strs = examples
        elif isinstance(examples, dict):
            for example_str, example_data in examples.items():
                if isinstance(example_data, str):
                    example_strs.append(example_str)

                elif isinstance(example_data, dict):

                    if (
                        example_data.get('runnable', True)
                        and not example_data.get('long', False)
                        and not example_data.get('skip', False)
                    ):
                        example_strs.append(example_str)

        # convert examples into raw commands
        for example_str in example_strs:
            command_pieces = [sys.executable, '-m', 'ctc'] + list(subcommand)

            # handle internal quotes
            if '\'' in example_str:
                pieces = example_str.split('\'')
                for head, tail in zip(pieces[::2], pieces[1::2]):
                    command_pieces.extend(head.split(' '))
                    command_pieces.append(tail)
            else:
                command_pieces.extend(example_str.split(' '))

            command_pieces = [
                piece.strip('"') for piece in command_pieces if piece != ''
            ]

            cli_examples.append(command_pieces)

    return cli_examples


cli_examples = collect_cli_examples()


@pytest.mark.parametrize('cli_example', cli_examples)
def test_subcommand_examples(cli_example):

    max_example_time = 30

    exit_code = subprocess.call(cli_example, timeout=max_example_time)
    if exit_code != 0:
        raise Exception(
            'command failed with exit code '
            + str(exit_code)
            + ': '
            + ' '.join(cli_example)
        )



from __future__ import annotations

from ctc import evm
from ctc import rpc
from ctc import spec

from . import ens_directory
from . import resolver


async def async_get_owner(
    name: str,
    *,
    block: spec.BlockNumberReference | None = None,
    context: spec.Context = None,
) -> str:
    node = resolver.hash_name(name)
    function_abi: spec.FunctionABI = {
        'name': 'owner',
        'inputs': [{'type': 'bytes32'}],
        'outputs': [{'type': 'address'}],
    }
    result = await rpc.async_eth_call(
        to_address=ens_directory.registry,
        function_abi=function_abi,
        function_parameters=[node],
        context=context,
        block_number=block,
    )
    if not isinstance(result, str):
        raise Exception('invalid rpc result')
    return result


async def async_record_exists(
    name: str,
    *,
    block: spec.BlockNumberReference | None = None,
    context: spec.Context = None,
) -> bool:
    node = resolver.hash_name(name)
    function_abi: spec.FunctionABI = {
        'name': 'recordExists',
        'inputs': [{'type': 'bytes32'}],
        'outputs': [{'type': 'bool'}],
    }
    result = await rpc.async_eth_call(
        to_address=ens_directory.registry,
        function_abi=function_abi,
        function_parameters=[node],
        context=context,
        block_number=block,
    )
    if not isinstance(result, bool):
        raise Exception('invalid rpc result')
    return result


async def async_get_resolver(
    name: str,
    *,
    block: spec.BlockNumberReference | None = None,
    context: spec.Context = None,
) -> spec.Address:
    node = resolver.hash_name(name)
    function_abi: spec.FunctionABI = {
        'name': 'resolver',
        'inputs': [{'type': 'bytes32'}],
        'outputs': [{'type': 'address'}],
    }
    result = await rpc.async_eth_call(
        to_address=ens_directory.registry,
        function_abi=function_abi,
        function_parameters=[node],
        context=context,
        block_number=block,
    )
    if not isinstance(result, str):
        raise Exception('invalid rpc result')
    return result


async def async_get_registration_block(
    name: str, *, context: spec.Context = None
) -> int:

    if not name.endswith('.eth'):
        raise NotImplementedError()

    label, *parent = name.split('.')
    parent_node = resolver.hash_name('.'.join(parent))

    registrations: spec.DataFrame = await async_get_registrations(
        context=context
    )
    mask = (registrations['arg__label'] == evm.keccak_text(label)) & (
        registrations['arg__parent_node'] == parent_node
    )
    result = registrations[mask]

    if len(result) == 0:
        raise Exception('could not find registration')

    block = registrations['block_number'][0]

    if not isinstance(block, int):
        raise Exception('invalid rpc result')

    return block


async def async_get_registrations(
    *,
    context: spec.Context = None,
) -> spec.DataFrame:
    event_abi: spec.EventABI = {
        'name': 'NewOwner',
        'inputs': [
            {
                'indexed': True,
                'name': 'node',
                'type': 'bytes32',
            },
            {
                'indexed': True,
                'name': 'label',
                'type': 'bytes32',
            },
            {
                'indexed': False,
                'name': 'owner',
                'type': 'address',
            },
        ],
    }

    new_owners = await evm.async_get_events(
        contract_address='0x00000000000c2e074ec69a0dfb2997ba6c7d2e1e',
        event_abi=event_abi,
        start_block=9000000,
        verbose=False,
        context=context,
    )
    new_owners = new_owners.rename({'arg__node': 'arg__parent_node'})
    return new_owners




from __future__ import annotations

import time
import typing

from ctc import evm
from ctc import spec
from . import block_convert

if typing.TYPE_CHECKING:
    import asyncio
    from typing_extensions import TypedDict

    class LatestBlockCacheEntry(TypedDict, total=False):
        request_time: float
        response_time: float
        block_number: int


async def async_get_block(
    block: spec.BlockReference,
    *,
    context: spec.Context = None,
) -> spec.DBBlock:
    """get block from local database or from RPC node"""

    if spec.is_block_number_reference(block):
        return await _async_get_block_by_number(block=block, context=context)
    elif spec.is_block_hash(block):
        from ctc import rpc

        rpc_block_data = await rpc.async_eth_get_block_by_hash(
            block_hash=block,
            context=context,
            include_full_transactions=False,
        )
        return block_convert.convert_rpc_block_to_db_block(rpc_block_data)
    else:
        raise Exception('unknown block specifier: ' + str(block))


async def _async_get_block_by_number(
    block: spec.BlockNumberReference,
    *,
    context: spec.Context = None,
) -> spec.DBBlock:

    from ctc import config
    from ctc import db
    from ctc import rpc

    read_cache, write_cache = config.get_context_cache_read_write(
        schema_name='blocks', context=context
    )

    if read_cache and isinstance(block, int):
        db_block_data = await db.async_query_block(
            block_number=block,
            context=context,
        )
        if db_block_data is not None:
            return db_block_data

    rpc_block_data: spec.RPCBlock = await rpc.async_eth_get_block_by_number(
        block_number=evm.standardize_block_number(block),
        context=context,
        include_full_transactions=False,
    )
    db_block = block_convert.convert_rpc_block_to_db_block(rpc_block_data)

    if write_cache:
        await db.async_intake_block(
            db_block=db_block,
            context=context,
        )

    return db_block


#
# # plural
#


async def async_get_blocks(
    blocks: typing.Sequence[spec.BlockReference],
    *,
    latest_block_number: int | None = None,
    context: spec.Context = None,
) -> typing.Sequence[spec.DBBlock]:
    """get blocks from local database or from RPC node"""

    from ctc import rpc

    if all(spec.is_block_number_reference(block) for block in blocks):
        return await _async_get_blocks_by_numbers(
            blocks=blocks,
            latest_block_number=latest_block_number,
            context=context,
        )
    elif all(spec.is_block_hash(block) for block in blocks):
        return await rpc.async_batch_eth_get_block_by_hash(
            block_hashes=blocks,
            include_full_transactions=False,
            context=context,
        )
    else:
        raise Exception(
            'blocks should be all block number references or all block hashes'
        )


async def _async_get_blocks_by_numbers(
    blocks: typing.Sequence[spec.BlockReference],
    *,
    latest_block_number: int | None = None,
    context: spec.Context = None,
) -> typing.Sequence[spec.DBBlock]:

    from ctc import config
    from ctc import rpc

    standardized = [evm.standardize_block_number(block) for block in blocks]
    pending = standardized

    read_cache, write_cache = config.get_context_cache_read_write(
        schema_name='blocks', context=context
    )

    if read_cache:
        from ctc import db

        db_block_datas = await db.async_query_blocks(
            block_numbers=pending, context=context
        )
        if db_block_datas is None:
            block_data_map = {}
        else:
            block_data_map = dict(zip(pending, db_block_datas))
            pending = [
                block
                for block, db_block_data in block_data_map.items()
                if db_block_data is None
            ]
    else:
        block_data_map = {}

    rpc_blocks = await rpc.async_batch_eth_get_block_by_number(
        block_numbers=pending,
        include_full_transactions=False,
        context=context,
    )
    db_blocks = [
        block_convert.convert_rpc_block_to_db_block(block)
        for block in rpc_blocks
    ]

    # intake rpc data to db
    if write_cache:
        from ctc import db

        await db.async_intake_blocks(
            db_blocks=db_blocks,
            latest_block_number=latest_block_number,
            context=context,
        )

    block_data_map.update(dict(zip(pending, db_blocks)))
    blocks_data = [block_data_map[block] for block in standardized]

    return blocks_data  # type: ignore


#
# # latest block number
#

_latest_block_cache: typing.MutableMapping[int, LatestBlockCacheEntry] = {}
_latest_block_lock: typing.MutableMapping[str, asyncio.Lock | None] = {
    'lock': None
}


async def async_get_latest_block_number(
    *,
    context: spec.Context = None,
    use_cache: bool = True,
    cache_time: int | float = 1,
) -> int:
    """get latest block number

    uses a per-network in-memory cache with a ttl of cache_time
    """

    from ctc import config
    from ctc import rpc

    if not use_cache:
        result = await rpc.async_eth_block_number(context=context)
        if not isinstance(result, int):
            raise Exception('invalid rpc result')
        return result

    else:

        # must initialize asyncio.Lock within a running event loop
        # see https://stackoverflow.com/a/55918049
        lock = _latest_block_lock['lock']
        if lock is None:
            import asyncio

            lock = asyncio.Lock()
            _latest_block_lock['lock'] = lock

        async with lock:

            network = config.get_context_chain_id(context)
            request_time = time.time()
            network_cache = _latest_block_cache.get(network)
            if (
                network_cache is not None
                and request_time - network_cache['request_time'] < cache_time
            ):
                return network_cache['block_number']

            result = await rpc.async_eth_block_number(context=context)
            if not isinstance(result, int):
                raise Exception('invalid rpc result')

            response_time = time.time()
            _latest_block_cache[network] = {
                'request_time': request_time,
                'response_time': response_time,
                'block_number': result,
            }

            return result


def sync_get_latest_block_number(
    *,
    context: spec.Context = None,
    use_cache: bool = True,
    cache_time: int | float = 1,
) -> int:
    """get latest block number"""
    from ctc import rpc

    result = rpc.sync_eth_block_number(context=context)
    if not isinstance(result, int):
        raise Exception('invalid rpc result')
    return result




from __future__ import annotations

from ... import query_utils
from . import block_gas_statements


async_query_median_block_gas_fee = query_utils.wrap_selector_with_connection(
    block_gas_statements.async_select_median_block_gas_fee,
    'block_gas',
)

async_query_median_blocks_gas_fees = query_utils.wrap_selector_with_connection(
    block_gas_statements.async_select_median_blocks_gas_fees,
    'block_gas',
)



from __future__ import annotations

import toolcli

from ctc import evm


def get_command_spec() -> toolcli.CommandSpec:
    return {
        'f': selector_command,
        'help': '''compute 4byte function selector using function signature

in most cli shells must enclose parentheses in quotes " "''',
        'args': [
            {'name': 'text_signature', 'help': 'signature of function'},
        ],
        'examples': ['"totalSupply()"', '"transfer(address,uint256)"'],
    }


def selector_command(text_signature: str) -> None:
    if '(' not in text_signature and ')' not in text_signature:
        text_signature = text_signature + '()'
    selector = evm.keccak_text(text_signature)[:10]
    print(selector)



from __future__ import annotations

import ast
import toolcli

from ctc import evm


def get_command_spec() -> toolcli.CommandSpec:
    return {
        'f': encode_command,
        'help': """encode data as EVM datatypes

for simple datatypes, no quotes required

for nested datatypes, enclose in quotes and quote contained addresses""",
        'args': [
            {
                'name': 'type',
                'help': 'datatype used for encoding',
                # 'dest': 'datatype',
            },
            {'name': 'data', 'help': 'data to be encoded'},
            {
                'name': '--packed',
                'help': 'encode like solidity\'s abi.encodePacked()',
                'action': 'store_true',
            },
        ],
        'examples': [
            'address 0x6b175474e89094c44da98b954eedeac495271d0f',
            '"(int64,int64,int64)" "[1,2,3]"',
        ],
    }


def encode_command(*, type: str, data: str, packed: bool) -> None:

    if ',' in type:
        if not type.startswith('('):
            type = '(' + type + ')'

    if data.startswith('0x'):
        literal_data = data
    else:
        try:
            literal_data = ast.literal_eval(data)
        except Exception:
            literal_data = ast.literal_eval('"' + data + '"')

    if packed:
        encoded = evm.abi_encode_packed(literal_data, type)
    else:
        encoded = evm.abi_encode(literal_data, type)

    as_hex = evm.to_hex(encoded)
    print(as_hex)



from __future__ import annotations

import typing

from ctc import rpc
from ctc import spec

from .. import chainlink_feed_metadata
from .. import chainlink_spec


# overloading in python is a mess...
@typing.overload
async def async_get_feed_datum(
    feed: chainlink_spec._FeedReference,
    *,
    fields: typing.Literal['full'],
    normalize: bool = True,
    invert: bool = False,
    block: typing.Optional[spec.BlockNumberReference] = None,
    context: spec.Context = None,
) -> chainlink_spec.FeedRoundData:
    ...


@typing.overload
async def async_get_feed_datum(
    feed: chainlink_spec._FeedReference,
    *,
    fields: typing.Literal['answer'] = 'answer',
    normalize: bool = True,
    invert: bool = False,
    block: typing.Optional[spec.BlockNumberReference] = None,
    context: spec.Context = None,
) -> typing.Union[int, float]:
    ...


async def async_get_feed_datum(
    feed: chainlink_spec._FeedReference,
    *,
    fields: typing.Literal['answer', 'full'] = 'answer',
    normalize: bool = True,
    invert: bool = False,
    block: typing.Optional[spec.BlockNumberReference] = None,
    context: spec.Context = None,
) -> typing.Union[int, float, chainlink_spec.FeedRoundData]:
    """get feed data for a single block"""

    if block is None:
        block = 'latest'

    feed = await chainlink_feed_metadata.async_resolve_feed_address(
        feed, context=context
    )

    if fields == 'answer':

        result = await rpc.async_eth_call(
            to_address=feed,
            function_abi=chainlink_spec.feed_function_abis['latestAnswer'],
            block_number=block,
            fill_empty=True,
            empty_token=None,
            context=context,
        )

        if not isinstance(result, (int, float)):
            raise Exception('invalid rpc result')
        answer = result

        if normalize:
            decimals = await chainlink_feed_metadata.async_get_feed_decimals(
                feed,
                context=context,
            )
            answer /= 10**decimals

        if invert:
            answer = 1 / answer

        return answer

    elif fields == 'full':

        data = await rpc.async_eth_call(
            to_address=feed,
            function_abi=chainlink_spec.feed_function_abis['latestRoundData'],
            block_number=block,
            fill_empty=True,
            empty_token=None,
            context=context,
        )

        round_id, answer, _started_at, updated_at, _answered_in_round_id = data

        full: chainlink_spec.FeedRoundData = {
            'answer': answer,
            'timestamp': updated_at,
            'round_id': round_id,
        }

        if answer is not None:
            if normalize:
                decimals = (
                    await chainlink_feed_metadata.async_get_feed_decimals(
                        feed, context=context
                    )
                )
                full['answer'] /= 10**decimals

            if invert:
                full['answer'] = 1 / full['answer']

            return full

        else:

            return full

    else:
        raise Exception('unknown fields type: ' + str(fields))




from __future__ import annotations

import toolcli

from ctc import evm


def get_command_spec() -> toolcli.CommandSpec:
    return {
        'f': integer_command,
        'help': 'convert data to integer',
        'args': [
            {'name': 'data', 'help': 'hex data to convert to integer'},
        ],
        'examples': [
            '0xa00',
            '0x123456789abdef',
        ],
    }


def integer_command(data: str) -> None:
    as_int = evm.binary_convert(data, 'integer')
    print(as_int)



import pytest

from ctc import evm


example_domains = {
    '0x2901a982e363189e3f2e4db2e5c3291fa1067b815a3ac9890ac6573e51bf33b0': {
        'name': 'Aave Token',
        'version': '1',
        'chainId': 1,
        'verifyingContract': '0x7Fc66500c84A76Ad7e9c93437bFc5Ac33E2DDaE9',
    },
    '0x0ff37d1dc9a7d6539faa7634e8848d1d0bb7ec77e2bcb30d8982e62e6e2d5914': {
        'name': 'Fei USD',
        'version': '1',
        'chainId': 1,
        'verifyingContract': '0x956f47f50a910163d8bf957cf5846d573e7f87ca',
    },
}


@pytest.mark.parametrize('test', example_domains.items())
def test_get_domain_separator(test):
    target_domain_separator, domain = test
    actual_domain_separator = evm.get_domain_separator(domain)
    assert target_domain_separator == actual_domain_separator



from __future__ import annotations

import asyncio
import typing

import tooltime

from ctc import evm
from ctc import spec
from . import twap_data
from . import twap_filter
from . import twap_spec


async def async_get_twap(
    data_source: twap_spec.DataSource,
    #
    # # data samples
    filter_duration: tooltime.Timestamp,
    *,
    #
    # # output samples
    output_start_block: typing.Optional[spec.BlockNumberReference] = None,
    output_end_block: typing.Optional[spec.BlockNumberReference] = None,
    output_start_time: typing.Optional[tooltime.Timestamp] = None,
    output_end_time: typing.Optional[tooltime.Timestamp] = None,
    #
    # # other
    context: spec.Context = None,
) -> spec.DataFrame:

    # get block numbers and timestamps
    start_coroutine = evm.async_get_block_number_and_time(
        block_number=output_start_block,
        block_timestamp=output_start_time,
        context=context,
    )
    start_task = asyncio.create_task(start_coroutine)
    end_coroutine = evm.async_get_block_number_and_time(
        block_number=output_end_block,
        block_timestamp=output_end_time,
        context=context,
    )
    end_task = asyncio.create_task(end_coroutine)
    start_block, start_time = await start_task
    end_block, end_time = await end_task

    # start block timestamp acquisition task
    blocks = list(range(start_block, end_block + 1))
    block_timestamps_task = asyncio.create_task(
        evm.async_get_block_timestamps(blocks=blocks, context=context)
    )

    # get data
    data = await twap_data.async_get_data_feed(
        data_source=data_source,
        start_block=start_block,
        end_block=end_block,
    )

    # filter data if necessary
    mode = data_source.get('mode')
    if mode == 'native':
        twap = data
    elif mode == 'raw':
        twap = twap_filter.filter_twap(
            raw_values=data.rows(),
            timestamps=(await block_timestamps_task),
            filter_duration=filter_duration,
        )
    else:
        raise Exception('unknown mode: ' + str(mode))

    return twap


async def async_get_twap_single_sample(
    data_source: twap_spec.DataSource,
    filter_duration: tooltime.Timestamp,
    *,
    block: typing.Optional[spec.BlockNumberReference] = None,
    timestamp: typing.Optional[tooltime.Timestamp] = None,
    context: spec.Context = None,
) -> float:

    series = await async_get_twap(
        data_source=data_source,
        filter_duration=filter_duration,
        output_start_block=block,
        output_end_block=block,
        output_start_time=timestamp,
        output_end_time=timestamp,
        context=context,
    )

    if len(series) == 0:
        raise Exception()
    elif len(series) > 1:
        raise Exception()
    else:
        return typing.cast(float, series[0])



from __future__ import annotations

import toolcli

from ctc.cli.cli_utils import cli_alias_utils


aliases_help = """these aliases allow you to call ctc subcommands without typing the "ctc" part

    for example, you can type [option]4byte 0xa9059cbb[/option] instead of [option]ctc 4byte 0xa9059cbb[/option]

these aliases must go in your shell config file

    this can be done using one of the following methods:
    1. use [option]ctc aliases --append[/option] to append to autodetected config files
    2. use [option]ctc aliases >> CONFIG_PATH[/option] to append the aliases to config file
    3. manually paste the aliases into your config file

    the path to your terminal file depends on shell:
    - bash: [metavar]~/.profile[/metavar] or [metavar]~/.bashrc[/metavar]
    - zsh: [metavar]~/.zshrc[/metavar]
    - fish: [metavar]~/.config/fish/config.fish[/metavar]

to learn more about the underlying commands, run [option]ctc help[/option]

depending on preference, can use all aliases or just a subset of them"""


def get_command_spec() -> toolcli.CommandSpec:
    return {
        'f': aliases_command,
        'help': 'view and manage ctc shell aliases' + '\n\n' + aliases_help,
        'args': [
            {
                'name': '--raw',
                'help': 'output aliases as raw script file',
                'action': 'store_true',
            },
            {
                'name': '--install',
                'help': 'install/upgrade aliases in shell config (will ask for confirmation)',
                'action': 'store_true',
            },
        ],
        'examples': {
            '': {},
            '--raw': {},
            '--install': {'runnable': False},
        },
    }


def aliases_command(*, raw: bool, install: bool) -> None:

    if raw:
        cli_alias_utils.print_raw_aliases()
    elif install:
        cli_alias_utils.install_aliases()
    else:
        cli_alias_utils.print_alias_status()



from __future__ import annotations

import os
from ctc import spec


def get_etherscan_key(network: spec.NetworkName) -> str | None:
    if network == 'ethereum':
        key = os.environ.get('ETHERSCAN_API_KEY')
        if key == '':
            return None
        else:
            return key
    else:
        return None



from __future__ import annotations

import typing

from ctc import spec


def combine_contract_abis(
    contract_abis: typing.Sequence[spec.ContractABI],
) -> spec.ContractABI:
    """combine multiple contract ABI's into single contract ABI"""

    # insert using name to avoid name collisions
    combined = {}
    unnamed = None
    for contract_abi in contract_abis:
        for item in contract_abi:
            if item.get('name') is None:
                unnamed = item
            else:
                combined[item['name']] = item

    new_abi = list(combined.values())
    if unnamed is not None:
        new_abi.append(unnamed)

    return new_abi



from __future__ import annotations

import os

import toolcli
import toolstr
import tooltime

from ctc import cli
from ctc import evm
from ctc import spec
from ctc.cli import cli_utils
from ctc.defi import dex_utils


def get_command_spec() -> toolcli.CommandSpec:
    return {
        'f': async_trades_command,
        'help': 'get DEX swaps',
        'args': [
            {'name': 'pool', 'help': 'pool to retrieve swaps from'},
            {
                'name': ['-b', '--blocks'],
                'help': 'block range to retrieve swaps from',
            },
            {
                'name': '--no-normalize',
                'help': 'do not normalize by ERC20 decimals',
            },
            {
                'name': '--export',
                'help': 'file path to save output to',
            },
            {
                'name': '--overwrite',
                'action': 'store_true',
                'help': 'specify that output path can be overwritten',
            },
            {
                'name': ['-v', '--verbose'],
                'action': 'store_true',
                'help': 'output extra data',
            },
        ],
        'examples': [
            '0xc5be99a02c6857f9eac67bbce58df5572498f40c --blocks 14000000:14010000',
        ],
    }


async def async_trades_command(
    *,
    pool: spec.Address,
    blocks: str,
    no_normalize: bool,
    export: str,
    overwrite: bool,
    verbose: bool,
) -> None:

    import polars as pl

    if blocks is not None:
        start_block, end_block = await cli_utils.async_parse_block_range(blocks)
    else:
        start_block = None
        end_block = None

    trades = await dex_utils.async_get_pool_trades(
        pool=pool,
        start_block=start_block,
        end_block=end_block,
        label='symbol',
        normalize=not no_normalize,
        include_prices=True,
        include_volumes=True,
    )

    if export is not None:
        if not overwrite and os.path.isfile(export):
            raise Exception('file already exists, use --overwrite')
        if export.endswith('.csv'):
            trades.write_csv(export)
        elif export.endswith('.json'):
            trades.write_json(export, row_oriented=True)
        else:
            raise Exception('unknown export format')
    else:

        trades = trades[-100:]

        block_timestamps = await evm.async_get_block_timestamps(
            trades['block_number']
        )
        timestamp_list = [
            tooltime.timestamp_to_iso_pretty(timestamp)
            for timestamp in block_timestamps
        ]
        trades = trades.with_columns(pl.Series('timestamp', timestamp_list))

        rename_columns = {
            key: key.replace('__', '\n') for key in list(trades.columns)
        }
        rename_columns['block_number'] = 'block'
        rename_columns['sold_id'] = 'sold'
        rename_columns['bought_id'] = 'bought'
        for key, value in rename_columns.items():
            if value.startswith('price\n'):
                rename_columns[key] = value[6:]
        trades = trades.rename(rename_columns)

        columns = list(trades.columns)
        if not verbose:
            columns = [
                column
                for column in columns
                if column
                not in [
                    'transaction_hash',
                    'recipient',
                    'sold_amount',
                    'bought_amount',
                ]
            ]

        column_formats = {
            column: {'trailing_zeros': True, 'decimals': 2}
            for column in columns
        }
        column_formats['block'] = {'commas': False}

        styles = cli.get_cli_styles()
        column_styles = {
            'block': styles['metavar'],
            'sold': styles['option'],
            'bought': styles['option'],
        }
        for column in columns:
            if '\nper\n' in column or 'volume\n' in column:
                column_styles[column] = styles['description'] + ' bold'

        toolstr.print_dataframe_as_table(
            trades,
            columns=columns,
            compact=2,
            column_formats=column_formats,
            column_styles=column_styles,
            border=styles['comment'],
            include_index=False,
            label_style=styles['title'],
        )



from __future__ import annotations

# getPoolOwnership
# getMaxRedeem
# getMaxBorrow
# getRewardSpeedsByPool
# getRewardSpeedsByPools
# getUnclaimedRewardsByDistributors
# getRewardsDistributorsBySupplier



from __future__ import annotations

import typing

from ctc import evm
from ctc import spec


def parse_network(network: str) -> spec.NetworkReference:
    if network.isnumeric():
        return int(network)
    else:
        return network


async def async_parse_block(
    block: str, context: spec.Context = None
) -> spec.BlockNumberReference:
    if block.isnumeric():
        return int(block)
    elif block == 'latest':
        return block
    else:
        raise Exception('unknown block format: ' + str(block))


async def async_parse_block_range(
    text: str,
    *,
    default_start: int | None = None,
    default_end: int | None = None,
    context: spec.Context = None,
) -> tuple[int | None, int | None]:
    """convert a cli block range to its start and end bounds

    Example block slices
        16000:        -->   (16000, None)
        :16010        -->   (None, 16010)
        16000:16010   -->   (16000, 16010)
        14e6:15e6     -->   (14000000, 15000000)
    """

    if text.count(':') != 1:
        raise Exception('range must have format: [start_block]:[end_block]')

    raw_start_block, raw_end_block = text.split(':')

    if raw_start_block == '':
        start_block = default_start
    else:
        start_block = await _async_resolve_single_block(
            raw_start_block, context=context
        )

    if raw_end_block == '':
        end_block = default_end
    else:
        end_block = await _async_resolve_single_block(
            raw_end_block, context=context
        )

    return start_block, end_block


async def async_parse_block_chunks(
    text: str,
    *,
    default_interval: int | None = None,
    context: spec.Context = None,
) -> tuple[int, int, int]:
    """

    Examples
    14_000_000:14_000_100:10
    14M:14.1M:1K
    10:100:50
    14M:latest:500k
    """
    if text.count(':') == 2:
        start_block, end_block, interval = text.split(':')
    elif text.count(':') == 1:
        start_block, end_block = text.split(':')
        interval = None
    else:
        raise Exception(
            'block chunks must be specified as start_block:end_block:chunk_size'
        )
    start_block_int = await _async_resolve_single_block(
        start_block, context=context
    )
    end_block_int = await _async_resolve_single_block(
        end_block, context=context
    )
    if interval is not None:
        interval_int = await _async_resolve_single_block(
            interval, context=context
        )
    else:
        if default_interval is None:
            raise Exception('must specify default_interval')
        interval_int = default_interval
    return start_block_int, end_block_int, interval_int


def sync_parse_block_chunks(
    text: str,
    *,
    default_interval: int | None = None,
    context: spec.Context = None,
) -> tuple[int, int, int]:
    """

    Examples
    14_000_000:14_000_100:10
    14M:14.1M:1K
    10:100:50
    14M:latest:500k
    """
    if text.count(':') == 2:
        start_block, end_block, interval = text.split(':')
    elif text.count(':') == 1:
        start_block, end_block = text.split(':')
        interval = None
    else:
        raise Exception(
            'block chunks must be specified as start_block:end_block:chunk_size'
        )
    start_block_int = _sync_resolve_single_block(start_block, context=context)
    end_block_int = _sync_resolve_single_block(end_block, context=context)
    if interval is not None:
        interval_int = _sync_resolve_single_block(interval, context=context)
    else:
        if default_interval is None:
            raise Exception('must specify default_interval')
        interval_int = default_interval
    return start_block_int, end_block_int, interval_int


async def async_parse_block_slice(
    text: str | typing.Sequence[str],
    n: int | None = None,
    *,
    context: spec.Context = None,
) -> typing.Sequence[int]:
    """convert a cli block slice to a list of integer block numbers

    Example block slices
        16000
        16000 16010 16020 16030
        16010:16030:10
        16010:16030
        15000000:latest:100000
        16010:+20
        16030:-20
        latest:-20:2
        14e6:15e6
        14e6:15e6:1e5
    """

    # list of blocks
    if isinstance(text, (list, tuple)):
        blocks: list[int] = []
        for subtext in text:
            subblocks = await async_parse_block_slice(
                subtext, n=n, context=context
            )
            blocks.extend(subblocks)
        return blocks

    elif isinstance(text, str):
        # single block
        if text == 'latest':
            return [await evm.async_get_latest_block_number(context=context)]

        elif text.isnumeric():
            return [int(text)]

        # list of blocks
        elif ' ' in text:
            pieces = text.split(' ')
            blocks = []
            for piece in pieces:
                block = await _async_resolve_single_block(
                    piece, context=context
                )
                blocks.append(block)
            return blocks

        # block slice
        elif ':' in text:
            pieces = text.split(':')

            if len(pieces) == 2:
                raw_start_block = pieces[0]
                raw_end_block = pieces[1]
                interval: int | None = None
            elif len(pieces) == 3:
                raw_start_block = pieces[0]
                raw_end_block = pieces[1]
                interval = int(pieces[2])
            else:
                raise Exception('invalid block slice specification')

            if raw_end_block[0] in ['+', '-']:
                start_block = await _async_resolve_single_block(
                    raw_start_block, context=context
                )
                diff = await _async_resolve_single_block(
                    raw_end_block, context=context
                )
                end_block = start_block + diff
                if raw_end_block[0] == '-':
                    start_block, end_block = end_block, start_block
            else:
                start_block = await _async_resolve_single_block(
                    raw_start_block, context=context
                )
                end_block = await _async_resolve_single_block(
                    raw_end_block, context=context
                )

            if end_block < start_block:
                raise Exception('invalid block slice specification')

            if interval is None:
                if n is None:
                    interval = 1
                else:
                    interval = int((end_block - start_block) / n)

            blocks = list(range(start_block, end_block + 1, interval))
            if blocks[-1] != end_block:
                blocks.append(end_block)

            return blocks

    raise Exception('invalid block slice specification')


async def _async_resolve_single_block(
    text: str,
    context: spec.Context = None,
) -> int:
    chars = set(text)
    numbers = set('0123456789')

    if chars.issubset(numbers):
        return int(text)
    elif chars.issubset(numbers | {'_'}):
        import ast

        result: int = ast.literal_eval(text)
        return result
    elif len(text) > 1 and set(text[:-1]).issubset(numbers):
        if text[-1].lower() == 'b':
            factor = 1_000_000_000
        elif text[-1].lower() == 'm':
            factor = 1_000_000
        elif text[-1].lower() == 'k':
            factor = 1_000
        else:
            raise Exception('unknown suffix')
        return int(text[:-1]) * factor
    elif text == 'latest':
        return await evm.async_get_latest_block_number(context=context)
    else:
        # scentific notation
        try:
            as_float = float(text)
            as_int = int(as_float)
            if abs(as_float - as_int) > 0.00000001:
                raise Exception('must specify integer block values')
            return as_int
        except ValueError:
            raise Exception('could not parse block: ' + str(text))


def _sync_resolve_single_block(
    text: str,
    context: spec.Context = None,
) -> int:
    chars = set(text)
    numbers = set('0123456789')

    if chars.issubset(numbers):
        return int(text)
    elif chars.issubset(numbers | {'_'}):
        import ast

        result: int = ast.literal_eval(text)
        return result
    elif len(text) > 1 and set(text[:-1]).issubset(numbers):
        if text[-1].lower() == 'b':
            factor = 1_000_000_000
        elif text[-1].lower() == 'm':
            factor = 1_000_000
        elif text[-1].lower() == 'k':
            factor = 1_000
        else:
            raise Exception('unknown suffix')
        return int(text[:-1]) * factor
    elif text == 'latest':
        return evm.sync_get_latest_block_number(context=context)
    else:
        # scentific notation
        try:
            as_float = float(text)
            as_int = int(as_float)
            if abs(as_float - as_int) > 0.00000001:
                raise Exception('must specify integer block values')
            return as_int
        except ValueError:
            raise Exception('could not parse block: ' + str(text))




from __future__ import annotations

import typing

from typing_extensions import Literal, TypedDict, NotRequired

from . import address_types


StorageBackend = Literal['Filesystem']

FilesystemRoot = Literal['root', 'default']


class DexPool(TypedDict):
    address: address_types.Address
    factory: address_types.Address
    asset0: address_types.Address | None
    asset1: address_types.Address | None
    asset2: NotRequired[address_types.Address | None]
    asset3: NotRequired[address_types.Address | None]
    creation_block: int
    fee: int | None
    additional_data: NotRequired[typing.Mapping[typing.Any, typing.Any]]
    priority: NotRequired[int | None]



from __future__ import annotations

import toolcli

from ctc import evm


def get_command_spec() -> toolcli.CommandSpec:
    return {
        'f': create_address_command,
        'help': 'generate address in the style of CREATE or CREATE2',
        'args': [
            {'name': 'sender', 'help': 'sender address'},
            {
                'name': 'nonce_or_salt',
                'help': 'sender nonce (for CREATE) or salt (for CREATE2)',
            },
            {
                'name': 'init_code',
                'nargs': '?',
                'help': 'initialization bytecode for CREATE2',
            },
        ],
        'examples': [
            '0xd8da6bf26964af9d7eed9e03e53415d37aa96045 123',
            '0xd8da6bf26964af9d7eed9e03e53415d37aa96045 0xaabb 0xaabbccdd',
        ],
    }


def create_address_command(
    *,
    sender: str,
    nonce_or_salt: str,
    init_code: str,
) -> None:
    if init_code is not None:
        address = evm.get_created_address(
            sender=sender, salt=nonce_or_salt, init_code=init_code
        )
    else:
        address = evm.get_created_address(
            sender=sender, nonce=int(nonce_or_salt)
        )
    print(address)



from __future__ import annotations

import os
import typing

if typing.TYPE_CHECKING:
    import toolsql

    import mypy_extensions

import toolcli

import ctc
import ctc.config
from . import cli_utils


command_index_by_category: dict[str, toolcli.CommandIndex] = {
    'admin': {
        (): 'ctc.cli.commands.root_command',
        ('aliases',): 'ctc.cli.commands.admin.aliases_command',
        ('charset',): 'ctc.cli.commands.admin.charset_command',
        ('color',): 'ctc.cli.commands.admin.color_command',
        ('config',): 'ctc.cli.commands.admin.config_command',
        ('config', 'edit'): 'ctc.cli.commands.admin.config.edit_command',
        ('config', 'path'): 'ctc.cli.commands.admin.config.path_command',
        ('db',): 'ctc.cli.commands.admin.db.status_command',
        (
            'db',
            'create',
            'tables',
        ): 'ctc.cli.commands.admin.db.create_tables_command',
        ('db', 'drop'): 'ctc.cli.commands.admin.db.drop_command',
        ('db', 'login'): 'ctc.cli.commands.admin.db.login_command',
        ('log',): 'ctc.cli.commands.admin.log_command',
        ('setup',): 'ctc.cli.commands.admin.setup_command',
        ('chains',): 'ctc.cli.commands.admin.chains_command',
    },
    'compute': {
        ('ascii',): 'ctc.cli.commands.compute.ascii_command',
        (
            'create',
            'address',
        ): 'ctc.cli.commands.compute.create_address_command',
        ('checksum',): 'ctc.cli.commands.compute.checksum_command',
        ('decode', 'call'): 'ctc.cli.commands.compute.decode_call_command',
        ('decode',): 'ctc.cli.commands.compute.decode_command',
        ('encode',): 'ctc.cli.commands.compute.encode_command',
        ('hex',): 'ctc.cli.commands.compute.hex_command',
        ('int',): 'ctc.cli.commands.compute.integer_command',
        ('keccak',): 'ctc.cli.commands.compute.keccak_command',
        ('limits',): 'ctc.cli.commands.compute.limits_command',
        ('lower',): 'ctc.cli.commands.compute.lower_command',
        ('rlp', 'encode'): 'ctc.cli.commands.compute.rlp_encode',
        ('selector',): 'ctc.cli.commands.compute.selector_command',
    },
    'data': {
        ('abi',): 'ctc.cli.commands.data.abi_command',
        ('abi', 'diff'): 'ctc.cli.commands.data.abi_diff_command',
        ('address',): 'ctc.cli.commands.data.address_command',
        ('address', 'txs'): 'ctc.cli.commands.data.address_txs_command',
        ('bytecode',): 'ctc.cli.commands.data.bytecode_command',
        ('block',): 'ctc.cli.commands.data.block_command',
        ('blocks',): 'ctc.cli.commands.data.blocks_command',
        ('call',): 'ctc.cli.commands.data.call_command',
        ('call', 'all'): 'ctc.cli.commands.data.call_all_command',
        ('calls',): 'ctc.cli.commands.data.calls_command',
        ('chain',): 'ctc.cli.commands.data.chain_command',
        ('decompile',): 'ctc.cli.commands.data.decompile_command',
        ('dex', 'chart'): 'ctc.cli.commands.data.dex.chart_command',
        ('dex', 'pool'): 'ctc.cli.commands.data.dex.pool_command',
        ('dex', 'pools'): 'ctc.cli.commands.data.dex.pools_command',
        ('dex', 'trades'): 'ctc.cli.commands.data.dex.trades_command',
        ('eth', 'balance'): 'ctc.cli.commands.data.eth.balance_command',
        ('eth', 'balances'): 'ctc.cli.commands.data.eth.balances_command',
        ('erc20',): 'ctc.cli.commands.data.erc20_command',
        ('erc20', 'balance'): 'ctc.cli.commands.data.erc20.balance_command',
        ('erc20', 'balances'): 'ctc.cli.commands.data.erc20.balances_command',
        ('erc20', 'transfers'): 'ctc.cli.commands.data.erc20.transfers_command',
        ('events',): 'ctc.cli.commands.data.events_command',
        ('gas',): 'ctc.cli.commands.data.gas_command',
        ('proxy',): 'ctc.cli.commands.data.proxy_command',
        ('proxy', 'register'): 'ctc.cli.commands.data.proxy_register_command',
        ('storage',): 'ctc.cli.commands.data.storage_command',
        ('symbol',): 'ctc.cli.commands.data.symbol_command',
        ('timestamp',): 'ctc.cli.commands.data.timestamp_command',
        ('tx',): 'ctc.cli.commands.data.tx_command',
    },
    'protocol': {
        ('aave',): 'ctc.protocols.aave_v2_utils.cli.aave_command',
        (
            'aave',
            'addresses',
        ): 'ctc.protocols.aave_v2_utils.cli.aave_addresses_command',
        ('cg',): 'ctc.protocols.coingecko_utils.cli.cg_command',
        ('chainlink',): 'ctc.protocols.chainlink_utils.cli.chainlink_command',
        (
            'chainlink',
            'feeds',
        ): 'ctc.protocols.chainlink_utils.cli.chainlink_ls_command',
        ('curve', 'pools'): 'ctc.protocols.curve_utils.cli.curve_pools_command',
        ('ens',): 'ctc.protocols.ens_utils.cli.ens_command',
        ('ens', 'exists'): 'ctc.protocols.ens_utils.cli.ens.exists_command',
        ('ens', 'hash'): 'ctc.protocols.ens_utils.cli.ens.hash_command',
        ('ens', 'owner'): 'ctc.protocols.ens_utils.cli.ens.owner_command',
        ('ens', 'records'): 'ctc.protocols.ens_utils.cli.ens.records_command',
        ('ens', 'resolve'): 'ctc.protocols.ens_utils.cli.ens.resolve_command',
        ('ens', 'reverse'): 'ctc.protocols.ens_utils.cli.ens.reverse_command',
        ('etherscan',): 'ctc.protocols.etherscan_utils.cli.etherscan_command',
        ('gnosis',): 'ctc.protocols.gnosis_utils.cli.gnosis_command',
        ('llama',): 'ctc.protocols.llama_utils.cli.llama_command',
        ('llama', 'chain'): 'ctc.protocols.llama_utils.cli.llama_chain_command',
        (
            'llama',
            'chains',
        ): 'ctc.protocols.llama_utils.cli.llama_chains_command',
        (
            'llama',
            'protocol',
        ): 'ctc.protocols.llama_utils.cli.llama_protocol_command',
        (
            'llama',
            'protocols',
        ): 'ctc.protocols.llama_utils.cli.llama_protocols_command',
        ('llama', 'pool'): 'ctc.protocols.llama_utils.cli.llama_pool_command',
        ('llama', 'pools'): 'ctc.protocols.llama_utils.cli.llama_pools_command',
        ('rari',): 'ctc.protocols.rari_utils.cli.rari.fuse_command',
        ('rari', 'pools'): 'ctc.protocols.rari_utils.cli.rari.pools_command',
        (
            'uniswap',
            'chart',
        ): 'ctc.protocols.uniswap_v2_utils.cli.chart_command',
        (
            'uniswap',
            'mints',
        ): 'ctc.protocols.uniswap_v2_utils.cli.mints_command',
        (
            'uniswap',
            'burns',
        ): 'ctc.protocols.uniswap_v2_utils.cli.burns_command',
        (
            'uniswap',
            'swaps',
        ): 'ctc.protocols.uniswap_v2_utils.cli.swaps_command',
        ('uniswap', 'pool'): 'ctc.protocols.uniswap_v2_utils.cli.pool_command',
        ('yearn',): 'ctc.protocols.yearn_utils.cli.yearn_command',
        (
            'yearn',
            'addresses',
        ): 'ctc.protocols.yearn_utils.cli.yearn_addresses_command',
        ('4byte',): 'ctc.protocols.fourbyte_utils.cli.fourbyte_command',
        (
            '4byte',
            'build',
        ): 'ctc.protocols.fourbyte_utils.cli.fourbyte_build_command',
    },
}

command_index: typing.MutableMapping[
    toolcli.CommandSequence, toolcli.CommandSpecReference
] = {}
help_subcommand_categories = {}
for category, category_command_index in command_index_by_category.items():
    command_index.update(category_command_index)
    for subcommand in category_command_index.keys():
        help_subcommand_categories[subcommand] = category


description = """ctc is a tool for historical data analysis of Ethereum and other EVM chains

if using ctc for the first time, run:
    [option]ctc setup[/option]

for quick lookups of an address, block, tx, or timestamp, run:
    [option]ctc QUERY[/option]

    where [option]QUERY[/option] is an address, block number, tx hash, or timestamp"""

# for example...
#     [description]print block summary[/description]
#     [option]ctc 14000000[/option]

#     [description]print contract summary[/description]
#     [option]ctc 0x956f47f50a910163d8bf957cf5846d573e7f87ca[/option]

#     [description]print token summary[/description]
#     [option]ctc FEI[/option]

#     [description]print tx summary[/description]
#     [option]ctc 0x285117fc6ecd443fc2a39f72dd3df79e65d25c5d3d680d41f229e793a495ef99[/option]"""


cd_dir_help = {
    'code': 'directory where ctc code is stored',
    'data': 'directory where ctc stores data',
    'config': 'directory where ctc stores config',
}


def cd_dir_getter(dirname: str) -> str:
    import ctc
    import ctc.config

    if dirname == 'code':
        return ctc.__path__[0]

    elif dirname == 'data':
        return ctc.config.get_data_dir()

    elif dirname == 'config':
        return os.path.dirname(ctc.config.get_config_path())

    else:
        raise Exception('unknown directory: ' + str(dirname))


if os.environ.get('BUILDING_SPHINX') != '1':
    help_url_getter: None | typing.Callable[
        [
            mypy_extensions.NamedArg(tuple[str, ...], 'subcommand'),
            mypy_extensions.NamedArg(toolcli.ParseSpec, 'parse_spec'),
        ],
        str,
    ] = None
else:

    def help_url_getter(
        *,
        subcommand: typing.Tuple[str, ...],
        parse_spec: toolcli.ParseSpec,
    ) -> str:
        categories = parse_spec['config']['help_subcommand_categories']
        category = categories.get(subcommand, 'other')
        return (
            'https://ctc.readthedocs.io/en/latest/cli/subcommands/'
            + category
            + '/'
            + '_'.join(subcommand)
            + '.html'
        )


def _db_config_getter() -> toolsql.DBConfig | None:
    import ctc.config

    import warnings

    warnings.warn('full context not given for db_config')

    return ctc.config.get_context_db_config(
        context={}, schema_name='schema_versions'
    )


# def _is_root_help(raw_command):
#     import sys

#     for command in [raw_command, sys.argv]:
#         if command is None:
#             continue
#         if '--cd-destination-tempfile' in command:
#             index = command.index('--cd-destination-tempfile')
#             command = command[:index] + command[index + 2 :]
#         if all(item.startswith('-') for item in command[1:]):
#             return True
#     return False


# def _print_help_from_cache(command_index, help_cache_dir):
#     help_cache_path = get_help_dir_hash_path(command_index, help_cache_dir)
#     if os.path.isfile(help_cache_path):
#         with open(help_cache_path, 'r') as f:
#             contents = f.read()
#         print(contents, end='')
#         return True
#     else:
#         return False


# def get_help_dir_hash_path(command_index, help_cache_dir):
#     import hashlib

#     command_index_str = str(sorted(command_index.items()))
#     name_hash = hashlib.md5(command_index_str.encode()).hexdigest()
#     help_cache_path = os.path.join(help_cache_dir, name_hash)
#     return help_cache_path


def run_cli(
    raw_command: str | None = None,
    **toolcli_kwargs: typing.Any,
) -> None:

    import tempfile

    help_cache_dir = os.path.join(tempfile.gettempdir(), 'ctc', 'help_cache')

    # the goal of the below code would be to avoid having to import toolcli and
    # save 6 ms of startup time. but the problem with the below code is that
    # it produces a different cache key because the plugins have not been added
    #     if _is_root_help(raw_command):
    #         try:
    #             printed = _print_help_from_cache(
    #                 command_index, help_cache_dir
    #             )
    #             if printed:
    #                 return
    #         except Exception:
    #             pass

    styles = cli_utils.get_cli_styles()

    config: toolcli.CLIConfig = {
        #
        # metadata
        'base_command': 'ctc',
        'description': description,
        'version': ctc.__version__,
        'cd_dir_help': cd_dir_help,
        'cd_dir_getter': cd_dir_getter,
        'help_url_getter': help_url_getter,  # type: ignore
        'help_cache_dir': help_cache_dir,
        'help_subcommand_categories': help_subcommand_categories,
        'async_context_manager': cli_utils.AsyncContextManager,
        'extra_data': {'styles': styles},
        'style_theme': styles,
        #
        # standard subcommands and standard args
        'include_standard_subcommands': True,
        'include_debug_arg': True,
    }

    toolcli_kwargs = dict({'config': config}, **toolcli_kwargs)

    toolcli.run_cli(
        raw_command=raw_command,
        command_index=command_index,
        **toolcli_kwargs,
    )




from __future__ import annotations

import toolcli

from ctc import evm
from ctc import rpc
from ctc import spec


def get_command_spec() -> toolcli.CommandSpec:
    return {
        'f': async_storage_command,
        'help': 'get value of storage slot',
        'args': [
            {'name': 'contract_address', 'help': 'address of contract'},
            {'name': 'slot', 'help': 'address of storage slot'},
            {'name': '--block', 'help': 'block number'},
            {
                'name': '--type',
                'help': 'decode data using this datatype',
                'dest': 'datatype',
            },
        ],
        'examples': [
            '0x245cc372c84b3645bf0ffe6538620b04a217988b 0x0',
            '0x245cc372c84b3645bf0ffe6538620b04a217988b 0x0 --type address',
            '0x245cc372c84b3645bf0ffe6538620b04a217988b 0x0 --block 14000000',
        ],
    }


async def async_storage_command(
    *,
    contract_address: str,
    slot: str,
    block: spec.BlockNumberReference | None,
    datatype: str | None,
) -> None:

    contract_address = await evm.async_resolve_address(
        contract_address,
        block=block,
    )

    result = await rpc.async_eth_get_storage_at(
        contract_address,
        position=slot,
        block_number=block,
    )
    if datatype is None:
        print(result)
    else:
        as_bytes = evm.to_binary(result)
        decoded = evm.abi_decode(as_bytes, datatype)
        print(decoded)
